"""Aspirate in place command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from opentrons.hardware_control import HardwareControlAPI

from .pipetting_common import (
    PipetteIdMixin,
    AspirateVolumeMixin,
    FlowRateMixin,
    BaseLiquidHandlingResult,
)
from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate
from ..errors.exceptions import PipetteNotReadyToAspirateError

if TYPE_CHECKING:
    from ..execution import PipettingHandler
    from ..state import StateView


AspirateInPlaceCommandType = Literal["aspirateInPlace"]


class AspirateInPlaceParams(PipetteIdMixin, AspirateVolumeMixin, FlowRateMixin):
    """Payload required to aspirate in place."""

    pass


class AspirateInPlaceResult(BaseLiquidHandlingResult):
    """Result data from the execution of a AspirateInPlace command."""

    pass


class AspirateInPlaceImplementation(
    AbstractCommandImpl[AspirateInPlaceParams, AspirateInPlaceResult]
):
    """AspirateInPlace command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        hardware_api: HardwareControlAPI,
        state_view: StateView,
        **kwargs: object,
    ) -> None:
        self._pipetting = pipetting
        self._state_view = state_view
        self._hardware_api = hardware_api

    async def execute(self, params: AspirateInPlaceParams) -> AspirateInPlaceResult:
        """Aspirate without moving the pipette.

        Raises:
            TipNotAttachedError: if no tip is attached to the pipette.
            PipetteNotReadyToAspirateError: pipette plunger is not ready.
        """
        ready_to_aspirate = self._pipetting.get_is_ready_to_aspirate(
            pipette_id=params.pipetteId,
        )

        if not ready_to_aspirate:
            raise PipetteNotReadyToAspirateError(
                "Pipette cannot aspirate in place because of a previous blow out."
                " The first aspirate following a blow-out must be from a specific well"
                " so the plunger can be reset in a known safe position."
            )
        volume = await self._pipetting.aspirate_in_place(
            pipette_id=params.pipetteId, volume=params.volume, flow_rate=params.flowRate
        )

        return AspirateInPlaceResult(volume=volume)


class AspirateInPlace(BaseCommand[AspirateInPlaceParams, AspirateInPlaceResult]):
    """AspirateInPlace command model."""

    commandType: AspirateInPlaceCommandType = "aspirateInPlace"
    params: AspirateInPlaceParams
    result: Optional[AspirateInPlaceResult]

    _ImplementationCls: Type[
        AspirateInPlaceImplementation
    ] = AspirateInPlaceImplementation


class AspirateInPlaceCreate(BaseCommandCreate[AspirateInPlaceParams]):
    """AspirateInPlace command request model."""

    commandType: AspirateInPlaceCommandType = "aspirateInPlace"
    params: AspirateInPlaceParams

    _CommandCls: Type[AspirateInPlace] = AspirateInPlace



from abc import ABC, abstractmethod
from typing import Optional


class AbstractEmulator(ABC):
    """Interface of gcode line processing hardware emulator."""

    @abstractmethod
    def handle(self, line: str) -> Optional[str]:
        """Handle a command and return a response."""
        ...

    @staticmethod
    def get_terminator() -> bytes:
        """Get the command terminator for messages coming from PI."""
        return b"\r\n\r\n"

    @staticmethod
    def get_ack() -> bytes:
        """Get the command ack send to the PI."""
        return b"ok\r\nok\r\n"



from typing import Dict, Any
from typing_extensions import Protocol

from .types import ConfigType
from opentrons.hardware_control.types import HardwareFeatureFlags


class Configurable(Protocol[ConfigType]):
    """Protocol specifying hardware control configuration."""

    def get_config(self) -> ConfigType:
        """Get the robot's configuration object.

        :returns .RobotConfig: The object.
        """
        ...

    def set_config(self, config: ConfigType) -> None:
        """Replace the currently-loaded config"""
        ...

    @property
    def hardware_feature_flags(self) -> HardwareFeatureFlags:
        ...

    @hardware_feature_flags.setter
    def hardware_feature_flags(self, feature_flags: HardwareFeatureFlags) -> None:
        """Replace the currently-configured hardware feature flags."""
        ...

    @property
    def config(self) -> ConfigType:
        ...

    @config.setter
    def config(self, config: ConfigType) -> None:
        ...

    async def update_config(self, **kwargs: Dict[str, Any]) -> None:
        """Update values of the robot's configuration.

        `kwargs` should contain keys of the robot's configuration. For
        instance, `update_config(log_level='debug)` would change the API
        server log level to logging.DEBUG.

        Documentation on keys can be found in the documentation for RobotConfig.
        """
        ...



"""ProtocolEngine-based Protocol API implementation core."""
from typing_extensions import Final

from opentrons.protocols.api_support.types import APIVersion

from .protocol import ProtocolCore
from .instrument import InstrumentCore
from .labware import LabwareCore
from .module_core import ModuleCore
from .well import WellCore

ENGINE_CORE_API_VERSION: Final = APIVersion(2, 14)

__all__ = [
    "ENGINE_CORE_API_VERSION",
    "ProtocolCore",
    "InstrumentCore",
    "LabwareCore",
    "WellCore",
    "ModuleCore",
]



"""Input file value objects."""
from __future__ import annotations
from typing import BinaryIO, Protocol


class AbstractInputFile(Protocol):
    """An individual file to be read as part of a protocol.

    Properties:
        filename: The basename, including extension, of the file.
        file: A [file](https://docs.python.org/3/glossary.html#term-file-object)
            providing the contents of the protocol to be read.
    """

    filename: str
    file: BinaryIO



from __future__ import annotations

import logging
from typing import (
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Type,
    Union,
    Mapping,
    cast,
)

from opentrons_shared_data.labware.dev_types import LabwareDefinition
from opentrons_shared_data.pipette.dev_types import PipetteNameType

from opentrons.types import Mount, Location, DeckLocation, DeckSlotName, StagingSlotName
from opentrons.legacy_broker import LegacyBroker
from opentrons.hardware_control import SyncHardwareAPI
from opentrons.hardware_control.modules.types import MagneticBlockModel
from opentrons.commands import protocol_commands as cmds, types as cmd_types
from opentrons.commands.helpers import stringify_labware_movement_command
from opentrons.commands.publisher import CommandPublisher, publish, publish_context
from opentrons.protocols.api_support import instrument as instrument_support
from opentrons.protocols.api_support.deck_type import (
    NoTrashDefinedError,
    should_load_fixed_trash_labware_for_python_protocol,
    should_load_fixed_trash_area_for_python_protocol,
)
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import (
    AxisMaxSpeeds,
    requires_version,
    APIVersionError,
)

from ._types import OffDeckType
from .core.common import ModuleCore, LabwareCore, ProtocolCore
from .core.core_map import LoadedCoreMap
from .core.engine.module_core import NonConnectedModuleCore
from .core.module import (
    AbstractTemperatureModuleCore,
    AbstractMagneticModuleCore,
    AbstractThermocyclerCore,
    AbstractHeaterShakerCore,
    AbstractMagneticBlockCore,
)
from .core.engine import ENGINE_CORE_API_VERSION
from .core.legacy.legacy_protocol_core import LegacyProtocolCore

from . import validation
from ._liquid import Liquid
from ._trash_bin import TrashBin
from ._waste_chute import WasteChute
from .deck import Deck
from .instrument_context import InstrumentContext
from .labware import Labware
from .module_contexts import (
    MagneticModuleContext,
    TemperatureModuleContext,
    ThermocyclerContext,
    HeaterShakerContext,
    MagneticBlockContext,
    ModuleContext,
)


logger = logging.getLogger(__name__)


ModuleTypes = Union[
    TemperatureModuleContext,
    MagneticModuleContext,
    ThermocyclerContext,
    HeaterShakerContext,
    MagneticBlockContext,
]


class HardwareManager(NamedTuple):
    """Back. compat. wrapper for a removed class called `HardwareManager`.

    This interface will not be present in PAPIv3.
    """

    hardware: SyncHardwareAPI


class ProtocolContext(CommandPublisher):
    """A context for the state of a protocol.

    The ``ProtocolContext`` class provides the objects, attributes, and methods that
    allow you to configure and control the protocol.

    Methods generally fall into one of two categories.

      - They can change the state of the ``ProtocolContext`` object, such as adding
        pipettes, hardware modules, or labware to your protocol.
      - They can control the flow of a running protocol, such as pausing, displaying
        messages, or controlling built-in robot hardware like the ambient lighting.

    Do not instantiate a ``ProtocolContext`` directly.
    The ``run()`` function of your protocol does that for you.
    See the :ref:`Tutorial <run-function>` for more information.

    Use :py:meth:`opentrons.execute.get_protocol_api` to instantiate a ``ProtocolContext`` when
    using Jupyter Notebook. See :ref:`advanced-control`.

    .. versionadded:: 2.0

    """

    def __init__(
        self,
        api_version: APIVersion,
        core: ProtocolCore,
        broker: Optional[LegacyBroker] = None,
        core_map: Optional[LoadedCoreMap] = None,
        deck: Optional[Deck] = None,
        bundled_data: Optional[Dict[str, bytes]] = None,
    ) -> None:
        """Build a :py:class:`.ProtocolContext`.

        :param api_version: The API version to use.
        :param core: The protocol implementation core.
        :param labware_offset_provider: Where this protocol context and its child
                                        module contexts will get labware offsets from.
        :param broker: An optional command broker to link to. If not
                      specified, a dummy one is used.
        :param bundled_data: A dict mapping filenames to the contents of data
                             files. Can be used by the protocol, since it is
                             exposed as
                             :py:attr:`.ProtocolContext.bundled_data`
        """
        super().__init__(broker)
        self._api_version = api_version
        self._core = core
        self._core_map = core_map or LoadedCoreMap()
        self._deck = deck or Deck(
            protocol_core=core, core_map=self._core_map, api_version=api_version
        )

        # With the introduction of Extension mount type, this dict initializes to include
        # the extension mount, for both ot2 & 3. While it doesn't seem like it would
        # create an issue in the current PAPI context, it would be much safer to
        # only use mounts available on the robot.
        self._instruments: Dict[Mount, Optional[InstrumentContext]] = {
            mount: None for mount in Mount
        }
        self._bundled_data: Dict[str, bytes] = bundled_data or {}

        # With the addition of Movable Trashes and Waste Chute support, it is not necessary
        # to ensure that the list of "disposal locations", essentially the list of trashes,
        # is initialized correctly on protocols utilizing former API versions prior to 2.16
        # and also to ensure that any protocols after 2.16 initialize a Fixed Trash for OT-2
        # protocols so that no load trash bin behavior is required within the protocol itself.
        # Protocols prior to 2.16 expect the Fixed Trash to exist as a Labware object, while
        # protocols after 2.16 expect trash to exist as either a TrashBin or WasteChute object.

        self._load_fixed_trash()
        if should_load_fixed_trash_labware_for_python_protocol(self._api_version):
            self._core.append_disposal_location(self.fixed_trash)
        elif should_load_fixed_trash_area_for_python_protocol(
            self._api_version, self._core.robot_type
        ):
            _fixed_trash_trashbin = TrashBin(
                location=DeckSlotName.FIXED_TRASH, addressable_area_name="fixedTrash"
            )
            # We are just appending the fixed trash to the core's internal list here, not adding it to the engine via
            # the core, since that method works through the SyncClient and if called from here, will cause protocols
            # to deadlock. Instead, that method is called in protocol engine directly in create_protocol_context after
            # ProtocolContext is initialized.
            self._core.append_disposal_location(_fixed_trash_trashbin)

        self._commands: List[str] = []
        self._unsubscribe_commands: Optional[Callable[[], None]] = None
        self.clear_commands()

    @property
    @requires_version(2, 0)
    def api_version(self) -> APIVersion:
        """Return the API version specified for this protocol context.

        This value is set when the protocol context
        is initialized.

          - When the context is the argument of ``run()``, the ``"apiLevel"`` key of the
            :ref:`metadata <tutorial-metadata>` or :ref:`requirements
            <tutorial-requirements>` dictionary determines ``api_version``.
          - When the context is instantiated with
            :py:meth:`opentrons.execute.get_protocol_api` or
            :py:meth:`opentrons.simulate.get_protocol_api`, the value of its ``version``
            argument determines ``api_version``.

        It may be lower than the :ref:`maximum version <max-version>` supported by the
        robot software, which is accessible via the
        ``protocol_api.MAX_SUPPORTED_VERSION`` constant.
        """
        return self._api_version

    @property
    def _hw_manager(self) -> HardwareManager:
        # TODO (lc 01-05-2021) remove this once we have a more
        # user facing hardware control http api.
        logger.warning(
            "This function will be deprecated in later versions."
            "Please use with caution."
        )
        return HardwareManager(hardware=self._core.get_hardware())

    @property
    @requires_version(2, 0)
    def bundled_data(self) -> Dict[str, bytes]:
        """Accessor for data files bundled with this protocol, if any.

        This is a dictionary mapping the filenames of bundled datafiles to their
        contents. The filename keys are formatted with extensions but without paths. For
        example, a file stored in the bundle as ``data/mydata/aspirations.csv`` will
        have the key ``"aspirations.csv"``. The values are :py:class:`bytes` objects
        representing the contents of the files.
        """
        return self._bundled_data

    def cleanup(self) -> None:
        """Finalize and clean up the protocol context."""
        if self._unsubscribe_commands:
            self._unsubscribe_commands()
            self._unsubscribe_commands = None

    def __del__(self) -> None:
        if getattr(self, "_unsubscribe_commands", None):
            self._unsubscribe_commands()  # type: ignore

    @property
    @requires_version(2, 0)
    def max_speeds(self) -> AxisMaxSpeeds:
        """Per-axis speed limits for moving instruments.

        Changing values within this property sets the speed limit for each non-plunger
        axis of the robot. Note that this property only sets upper limits and can't
        exceed the physical speed limits of the movement system.

        This property is a dict mapping string names of axes to float values
        of maximum speeds in mm/s. To change a speed, set that axis's value. To
        reset an axis's speed to default, delete the entry for that axis
        or assign it to ``None``.

        See :ref:`axis_speed_limits` for examples.

        .. note::
            This property is not yet supported in API version 2.14 or higher.
        """
        if self._api_version >= ENGINE_CORE_API_VERSION:
            # TODO(mc, 2023-02-23): per-axis max speeds not yet supported on the engine
            # See https://opentrons.atlassian.net/browse/RCORE-373
            raise APIVersionError(
                "ProtocolContext.max_speeds is not supported at apiLevel 2.14 or higher."
                " Use a lower apiLevel or set speeds using InstrumentContext.default_speed"
                " or the per-method 'speed' argument."
            )

        return self._core.get_max_speeds()

    @requires_version(2, 0)
    def commands(self) -> List[str]:
        """Return the run log.

        This is a list of human-readable strings representing what's been done in the protocol so
        far. For example, "Aspirating 123 µL from well A1 of 96 well plate in slot 1."

        The exact format of these entries is not guaranteed. The format here may differ from other
        places that show the run log, such as the Opentrons App or touchscreen.
        """
        return self._commands

    @requires_version(2, 0)
    def clear_commands(self) -> None:
        self._commands.clear()
        if self._unsubscribe_commands:
            self._unsubscribe_commands()

        def on_command(message: cmd_types.CommandMessage) -> None:
            payload = message.get("payload")

            if payload is None:
                return

            text = payload.get("text")

            if text is None:
                return

            if message["$"] == "before":
                self._commands.append(text)

        self._unsubscribe_commands = self.broker.subscribe(
            cmd_types.COMMAND, on_command
        )

    @requires_version(2, 0)
    def is_simulating(self) -> bool:
        """Returns ``True`` if the protocol is running in simulation.

        Returns ``False`` if the protocol is running on actual hardware.

        You can evaluate the result of this method in an ``if`` statement to make your
        protocol behave differently in different environments. For example, you could
        refer to a data file on your computer when simulating and refer to a data file
        stored on the robot when not simulating.

        You can also use it to skip time-consuming aspects of your protocol. Most Python
        Protocol API methods, like :py:meth:`.delay`, are designed to evaluate
        instantaneously in simulation. But external methods, like those from the
        :py:mod:`time` module, will run at normal speed if not skipped.
        """
        return self._core.is_simulating()

    @requires_version(2, 0)
    def load_labware_from_definition(
        self,
        labware_def: "LabwareDefinition",
        location: Union[DeckLocation, OffDeckType],
        label: Optional[str] = None,
    ) -> Labware:
        """Specify the presence of a labware on the deck.

        This function loads the labware definition specified by ``labware_def``
        to the location specified by ``location``.

        :param labware_def: The labware's definition.
        :param location: The slot into which to load the labware,
                         such as ``1``, ``"1"``, or ``"D1"``. See :ref:`deck-slots`.
        :type location: int or str or :py:obj:`OFF_DECK`
        :param str label: An optional special name to give the labware. If specified,
            this is how the labware will appear in the run log, Labware Position
            Check, and elsewhere in the Opentrons App and on the touchscreen.
        """
        load_params = self._core.add_labware_definition(labware_def)

        return self.load_labware(
            load_name=load_params.load_name,
            namespace=load_params.namespace,
            version=load_params.version,
            location=location,
            label=label,
        )

    @requires_version(2, 0)
    def load_labware(
        self,
        load_name: str,
        location: Union[DeckLocation, OffDeckType],
        label: Optional[str] = None,
        namespace: Optional[str] = None,
        version: Optional[int] = None,
        adapter: Optional[str] = None,
    ) -> Labware:
        """Load a labware onto a location.

        For Opentrons-verified labware, this is a convenient way
        to collapse the two stages of labware initialization (creating
        the labware and adding it to the protocol) into one.

        This function returns the created and initialized labware for use
        later in the protocol.

        :param str load_name: A string to use for looking up a labware definition.
            You can find the ``load_name`` for any Opentrons-verified labware on the
            `Labware Library <https://labware.opentrons.com>`__.

        :param location: Either a :ref:`deck slot <deck-slots>`,
            like ``1``, ``"1"``, or ``"D1"``, or the special value :py:obj:`OFF_DECK`.

            .. versionchanged:: 2.15
                You can now specify a deck slot as a coordinate, like ``"D1"``.

        :type location: int or str or :py:obj:`OFF_DECK`

        :param str label: An optional special name to give the labware. If specified,
            this is how the labware will appear in the run log, Labware Position
            Check, and elsewhere in the Opentrons App and on the touchscreen.

        :param str namespace: The namespace that the labware definition belongs to.
            If unspecified, the API will automatically search two namespaces:

              - ``"opentrons"``, to load standard Opentrons labware definitions.
              - ``"custom_beta"``, to load custom labware definitions created with the
                `Custom Labware Creator <https://labware.opentrons.com/create>`__.

            You might need to specify an explicit ``namespace`` if you have a custom
            definition whose ``load_name`` is the same as an Opentrons-verified
            definition, and you want to explicitly choose one or the other.

        :param version: The version of the labware definition. You should normally
            leave this unspecified to let ``load_labware()`` choose a version
            automatically.
        :param adapter: An adapter to load the labware on top of. Accepts the same
            values as the ``load_name`` parameter of :py:meth:`.load_adapter`. The
            adapter will use the same namespace as the labware, and the API will
            choose the adapter's version automatically.

                        .. versionadded:: 2.15
        """
        if isinstance(location, OffDeckType) and self._api_version < APIVersion(2, 15):
            raise APIVersionError(
                "Loading a labware off-deck requires apiLevel 2.15 or higher."
            )

        load_name = validation.ensure_lowercase_name(load_name)
        load_location: Union[OffDeckType, DeckSlotName, StagingSlotName, LabwareCore]
        if adapter is not None:
            if self._api_version < APIVersion(2, 15):
                raise APIVersionError(
                    "Loading a labware on an adapter requires apiLevel 2.15 or higher."
                )
            loaded_adapter = self.load_adapter(
                load_name=adapter,
                location=location,
                namespace=namespace,
            )
            load_location = loaded_adapter._core
        elif isinstance(location, OffDeckType):
            load_location = location
        else:
            load_location = validation.ensure_and_convert_deck_slot(
                location, self._api_version, self._core.robot_type
            )

        labware_core = self._core.load_labware(
            load_name=load_name,
            location=load_location,
            label=label,
            namespace=namespace,
            version=version,
        )

        labware = Labware(
            core=labware_core,
            api_version=self._api_version,
            protocol_core=self._core,
            core_map=self._core_map,
        )
        self._core_map.add(labware_core, labware)

        return labware

    @requires_version(2, 0)
    def load_labware_by_name(
        self,
        load_name: str,
        location: DeckLocation,
        label: Optional[str] = None,
        namespace: Optional[str] = None,
        version: int = 1,
    ) -> Labware:
        """
        .. deprecated:: 2.0
            Use :py:meth:`load_labware` instead.
        """
        logger.warning("load_labware_by_name is deprecated. Use load_labware instead.")
        return self.load_labware(load_name, location, label, namespace, version)

    @requires_version(2, 15)
    def load_adapter_from_definition(
        self,
        adapter_def: "LabwareDefinition",
        location: Union[DeckLocation, OffDeckType],
    ) -> Labware:
        """Specify the presence of an adapter on the deck.

        This function loads the adapter definition specified by ``adapter_def``
        to the location specified by ``location``.

        :param adapter_def: The adapter's labware definition.
        :param location: The slot into which to load the labware,
                         such as ``1``, ``"1"``, or ``"D1"``. See :ref:`deck-slots`.
        :type location: int or str or :py:obj:`OFF_DECK`
        """
        load_params = self._core.add_labware_definition(adapter_def)

        return self.load_adapter(
            load_name=load_params.load_name,
            namespace=load_params.namespace,
            version=load_params.version,
            location=location,
        )

    @requires_version(2, 16)
    def load_trash_bin(self, location: DeckLocation) -> TrashBin:
        """Load a trash bin on the deck of a Flex.

        See :ref:`configure-trash-bin` for details.

        If you try to load a trash bin on an OT-2, the API will raise an error.

        :param location: The :ref:`deck slot <deck-slots>` where the trash bin is. The
            location can be any unoccupied slot in column 1 or 3.

            If you try to load a trash bin in column 2 or 4, the API will raise an error.
        """
        slot_name = validation.ensure_and_convert_deck_slot(
            location,
            api_version=self._api_version,
            robot_type=self._core.robot_type,
        )
        if not isinstance(slot_name, DeckSlotName):
            raise ValueError("Staging areas not permitted for trash bin.")
        addressable_area_name = validation.ensure_and_convert_trash_bin_location(
            location,
            api_version=self._api_version,
            robot_type=self._core.robot_type,
        )
        trash_bin = TrashBin(
            location=slot_name, addressable_area_name=addressable_area_name
        )
        self._core.add_disposal_location_to_engine(trash_bin)
        return trash_bin

    @requires_version(2, 16)
    def load_waste_chute(
        self,
    ) -> WasteChute:
        """Load the waste chute on the deck of a Flex.

        See :ref:`configure-waste-chute` for details, including the deck configuration
        variants of the waste chute.

        The deck plate adapter for the waste chute can only go in slot D3. If you try to
        load another item in slot D3 after loading the waste chute, or vice versa, the
        API will raise an error.
        """
        waste_chute = WasteChute()
        self._core.add_disposal_location_to_engine(waste_chute)
        return waste_chute

    @requires_version(2, 15)
    def load_adapter(
        self,
        load_name: str,
        location: Union[DeckLocation, OffDeckType],
        namespace: Optional[str] = None,
        version: Optional[int] = None,
    ) -> Labware:
        """Load an adapter onto a location.

        For adapters already defined by Opentrons, this is a convenient way
        to collapse the two stages of adapter initialization (creating
        the adapter and adding it to the protocol) into one.

        This function returns the created and initialized adapter for use
        later in the protocol.

        :param str load_name: A string to use for looking up a labware definition for the adapter.
            You can find the ``load_name`` for any standard adapter on the Opentrons
            `Labware Library <https://labware.opentrons.com>`_.

        :param location: Either a :ref:`deck slot <deck-slots>`,
            like ``1``, ``"1"``, or ``"D1"``, or the special value :py:obj:`OFF_DECK`.

        :type location: int or str or :py:obj:`OFF_DECK`

        :param str namespace: The namespace that the labware definition belongs to.
            If unspecified, the API will automatically search two namespaces:

              * ``"opentrons"``, to load standard Opentrons labware definitions.
              * ``"custom_beta"``, to load custom labware definitions created with the
                `Custom Labware Creator <https://labware.opentrons.com/create>`_.

            You might need to specify an explicit ``namespace`` if you have a custom
            definition whose ``load_name`` is the same as an Opentrons standard
            definition, and you want to explicitly choose one or the other.

        :param version: The version of the labware definition. You should normally
            leave this unspecified to let ``load_adapter()`` choose a version automatically.
        """
        load_name = validation.ensure_lowercase_name(load_name)
        load_location: Union[OffDeckType, DeckSlotName, StagingSlotName]
        if isinstance(location, OffDeckType):
            load_location = location
        else:
            load_location = validation.ensure_and_convert_deck_slot(
                location, self._api_version, self._core.robot_type
            )

        labware_core = self._core.load_adapter(
            load_name=load_name,
            location=load_location,
            namespace=namespace,
            version=version,
        )

        adapter = Labware(
            core=labware_core,
            api_version=self._api_version,
            protocol_core=self._core,
            core_map=self._core_map,
        )
        self._core_map.add(labware_core, adapter)

        return adapter

    # TODO(mm, 2023-06-07): Figure out what to do with this, now that the Flex has non-integer
    # slot names and labware can be stacked. https://opentrons.atlassian.net/browse/RLAB-354
    @property
    @requires_version(2, 0)
    def loaded_labwares(self) -> Dict[int, Labware]:
        """Get the labwares that have been loaded into the protocol context.

        Slots with nothing in them will not be present in the return value.

        .. note::

            If a module is present on the deck but no labware has been loaded
            into it with ``module.load_labware()``, there will
            be no entry for that slot in this value. That means you should not
            use ``loaded_labwares`` to determine if a slot is available or not,
            only to get a list of labwares. If you want a data structure of all
            objects on the deck regardless of type, use :py:attr:`deck`.


        :returns: Dict mapping deck slot number to labware, sorted in order of
                  the locations.
        """
        labware_cores = (
            (core.get_deck_slot(), core) for core in self._core.get_labware_cores()
        )

        return {
            slot.as_int(): self._core_map.get(core)
            for slot, core in labware_cores
            if slot is not None
        }

    # TODO (spp, 2022-12-14): https://opentrons.atlassian.net/browse/RLAB-237
    @requires_version(2, 15)
    def move_labware(
        self,
        labware: Labware,
        new_location: Union[
            DeckLocation, Labware, ModuleTypes, OffDeckType, WasteChute
        ],
        use_gripper: bool = False,
        pick_up_offset: Optional[Mapping[str, float]] = None,
        drop_offset: Optional[Mapping[str, float]] = None,
    ) -> None:
        """Move a loaded labware to a new location.

        See :ref:`moving-labware` for more details.

        :param labware: The labware to move. It should be a labware already loaded
                        using :py:meth:`load_labware`.

        :param new_location: Where to move the labware to. This is either:

                * A deck slot like ``1``, ``"1"``, or ``"D1"``. See :ref:`deck-slots`.
                * A hardware module that's already been loaded on the deck
                  with :py:meth:`load_module`.
                * A labware or adapter that's already been loaded on the deck
                  with :py:meth:`load_labware` or :py:meth:`load_adapter`.
                * The special constant :py:obj:`OFF_DECK`.

        :param use_gripper: Whether to use the Flex Gripper for this movement.

                * If ``True``, use the gripper to perform an automatic
                  movement. This will raise an error in an OT-2 protocol.
                * If ``False``, pause protocol execution until the user
                  performs the movement. Protocol execution remains paused until
                  the user presses **Confirm and resume**.

        Gripper-only parameters:

        :param pick_up_offset: Optional x, y, z vector offset to use when picking up labware.
        :param drop_offset: Optional x, y, z vector offset to use when dropping off labware.

        Before moving a labware to or from a hardware module, make sure that the labware's
        current and new locations are accessible, i.e., open the Thermocycler lid or
        open the Heater-Shaker's labware latch.
        """

        if not isinstance(labware, Labware):
            raise ValueError(
                f"Expected labware of type 'Labware' but got {type(labware)}."
            )

        location: Union[
            ModuleCore,
            LabwareCore,
            WasteChute,
            OffDeckType,
            DeckSlotName,
            StagingSlotName,
        ]
        if isinstance(new_location, (Labware, ModuleContext)):
            location = new_location._core
        elif isinstance(new_location, (OffDeckType, WasteChute)):
            location = new_location
        else:
            location = validation.ensure_and_convert_deck_slot(
                new_location, self._api_version, self._core.robot_type
            )

        _pick_up_offset = (
            validation.ensure_valid_labware_offset_vector(pick_up_offset)
            if pick_up_offset
            else None
        )
        _drop_offset = (
            validation.ensure_valid_labware_offset_vector(drop_offset)
            if drop_offset
            else None
        )
        with publish_context(
            broker=self.broker,
            command=cmds.move_labware(
                # This needs to be called from protocol context and not the command for import loop reasons
                text=stringify_labware_movement_command(
                    labware, new_location, use_gripper
                )
            ),
        ):
            self._core.move_labware(
                labware_core=labware._core,
                new_location=location,
                use_gripper=use_gripper,
                pause_for_manual_move=True,
                pick_up_offset=_pick_up_offset,
                drop_offset=_drop_offset,
            )

    @requires_version(2, 0)
    def load_module(
        self,
        module_name: str,
        location: Optional[DeckLocation] = None,
        configuration: Optional[str] = None,
    ) -> ModuleTypes:
        """Load a module onto the deck, given its name or model.

        This is the function to call to use a module in your protocol, like
        :py:meth:`load_instrument` is the method to call to use an instrument
        in your protocol. It returns the created and initialized module
        context, which will be a different class depending on the kind of
        module loaded.

        After loading modules, you can access a map of deck positions to loaded modules
        with :py:attr:`loaded_modules`.

        :param str module_name: The name or model of the module.
            See :ref:`available_modules` for possible values.

        :param location: The location of the module.

            This is usually the name or number of the slot on the deck where you
            will be placing the module, like ``1``, ``"1"``, or ``"D1"``. See :ref:`deck-slots`.

            The Thermocycler is only valid in one deck location.
            You don't have to specify a location when loading it, but if you do,
            it must be ``7``, ``"7"``, or ``"B1"``. See :ref:`thermocycler-module`.

            .. versionchanged:: 2.15
                You can now specify a deck slot as a coordinate, like ``"D1"``.

        :param configuration: Configure a Thermocycler to be in the ``semi`` position.
            This parameter does not work. Do not use it.

            .. versionchanged:: 2.14
                This parameter dangerously modified the protocol's geometry system,
                and it didn't function properly, so it was removed.

        :type location: str or int or None
        :returns: The loaded and initialized module---a
                  :py:class:`HeaterShakerContext`,
                  :py:class:`MagneticBlockContext`,
                  :py:class:`MagneticModuleContext`,
                  :py:class:`TemperatureModuleContext`, or
                  :py:class:`ThermocyclerContext`,
                  depending on what you requested with ``module_name``.

                  .. versionchanged:: 2.13
                    Added ``HeaterShakerContext`` return value.

                  .. versionchanged:: 2.15
                    Added ``MagneticBlockContext`` return value.
        """
        if configuration:
            if self._api_version < APIVersion(2, 4):
                raise APIVersionError(
                    f"You have specified API {self._api_version}, but you are"
                    "using Thermocycler parameters only available in 2.4"
                )
            if self._api_version >= ENGINE_CORE_API_VERSION:
                raise APIVersionError(
                    "The configuration parameter of load_module has been removed."
                )

        requested_model = validation.ensure_module_model(module_name)
        if isinstance(
            requested_model, MagneticBlockModel
        ) and self._api_version < APIVersion(2, 15):
            raise APIVersionError(
                f"Module of type {module_name} is only available in versions 2.15 and above."
            )

        deck_slot = (
            None
            if location is None
            else validation.ensure_and_convert_deck_slot(
                location, self._api_version, self._core.robot_type
            )
        )
        if isinstance(deck_slot, StagingSlotName):
            raise ValueError("Cannot load a module onto a staging slot.")

        module_core = self._core.load_module(
            model=requested_model,
            deck_slot=deck_slot,
            configuration=configuration,
        )

        module_context = _create_module_context(
            module_core=module_core,
            protocol_core=self._core,
            core_map=self._core_map,
            broker=self._broker,
            api_version=self._api_version,
        )

        self._core_map.add(module_core, module_context)

        return module_context

    # TODO(mm, 2023-06-07): Figure out what to do with this, now that the Flex has non-integer
    # slot names and labware can be stacked. https://opentrons.atlassian.net/browse/RLAB-354
    @property
    @requires_version(2, 0)
    def loaded_modules(self) -> Dict[int, ModuleTypes]:
        """Get the modules loaded into the protocol context.

        This is a map of deck positions to modules loaded by previous calls to
        :py:meth:`load_module`. It does not reflect what modules are actually attached
        to the robot. For example, if the robot has a Magnetic Module and a Temperature
        Module attached, but the protocol has only loaded the Temperature Module with
        :py:meth:`load_module`, only the Temperature Module will be included in
        ``loaded_modules``.

        :returns: Dict mapping slot name to module contexts. The elements may not be
            ordered by slot number.
        """
        return {
            core.get_deck_slot().as_int(): self._core_map.get(core)
            for core in self._core.get_module_cores()
        }

    @requires_version(2, 0)
    def load_instrument(
        self,
        instrument_name: str,
        mount: Union[Mount, str, None] = None,
        tip_racks: Optional[List[Labware]] = None,
        replace: bool = False,
    ) -> InstrumentContext:
        """Load a specific instrument for use in the protocol.

        When analyzing the protocol on the robot, instruments loaded with this method
        are compared against the instruments attached to the robot. You won't be able to
        start the protocol until the correct instruments are attached and calibrated.

        Currently, this method only loads pipettes. You do not need to load the Flex
        Gripper to use it in protocols. See :ref:`automatic-manual-moves`.

        :param str instrument_name: The instrument to load. See :ref:`new-pipette-models`
                                    for the valid values.
        :param mount: The mount where the instrument should be attached.
                      This can either be an instance of :py:class:`.types.Mount` or one
                      of the strings ``"left"`` or ``"right"``. When loading a Flex
                      96-Channel Pipette (``instrument_name="flex_96channel_1000"``),
                      you can leave this unspecified, since it always occupies both
                      mounts; if you do specify a value, it will be ignored.
        :type mount: types.Mount or str or ``None``
        :param tip_racks: A list of tip racks from which to pick tips when calling
                          :py:meth:`.InstrumentContext.pick_up_tip` without arguments.
        :type tip_racks: List[:py:class:`.Labware`]
        :param bool replace: If ``True``, replace the currently loaded instrument in
                             ``mount``, if any. This is intended for :ref:`advanced
                             control <advanced-control>` applications. You cannot
                             replace an instrument in the middle of a protocol being run
                             from the Opentrons App or touchscreen.
        """
        instrument_name = validation.ensure_lowercase_name(instrument_name)
        checked_instrument_name = validation.ensure_pipette_name(instrument_name)
        checked_mount = validation.ensure_mount_for_pipette(
            mount, checked_instrument_name
        )

        is_96_channel = checked_instrument_name == PipetteNameType.P1000_96

        tip_racks = tip_racks or []

        # TODO (tz, 9-12-23): move validation into PE
        on_right_mount = self._instruments[Mount.RIGHT]
        if is_96_channel and on_right_mount is not None:
            raise RuntimeError(
                f"Instrument already present on right:"
                f" {on_right_mount.name}. In order to load a 96-channel pipette, both mounts need to be available."
            )

        existing_instrument = self._instruments[checked_mount]
        if existing_instrument is not None and not replace:
            # TODO(mc, 2022-08-25): create specific exception type
            raise RuntimeError(
                f"Instrument already present on {checked_mount.name.lower()}:"
                f" {existing_instrument.name}"
            )

        logger.info(
            f"Loading {checked_instrument_name} on {checked_mount.name.lower()} mount"
        )

        instrument_core = self._core.load_instrument(
            instrument_name=checked_instrument_name,
            mount=checked_mount,
        )

        for tip_rack in tip_racks:
            instrument_support.validate_tiprack(
                instrument_name=instrument_core.get_pipette_name(),
                tip_rack=tip_rack,
                log=logger,
            )

        trash: Optional[Union[Labware, TrashBin]]
        try:
            trash = self.fixed_trash
        except (NoTrashDefinedError, APIVersionError):
            trash = None

        instrument = InstrumentContext(
            core=instrument_core,
            protocol_core=self._core,
            broker=self._broker,
            api_version=self._api_version,
            tip_racks=tip_racks,
            trash=trash,
            requested_as=instrument_name,
        )

        self._instruments[checked_mount] = instrument

        return instrument

    @property
    @requires_version(2, 0)
    def loaded_instruments(self) -> Dict[str, InstrumentContext]:
        """Get the instruments that have been loaded into the protocol.

        This is a map of mount name to instruments previously loaded with
        :py:meth:`load_instrument`. It does not reflect what instruments are actually
        installed on the robot. For example, if the robot has instruments installed on
        both mounts but your protocol has only loaded one of them with
        :py:meth:`load_instrument`, the unused one will not be included in
        ``loaded_instruments``.

        :returns: A dict mapping mount name (``"left"`` or ``"right"``) to the
            instrument in that mount. If a mount has no loaded instrument, that key
            will be missing from the dict.
        """
        return {
            mount.name.lower(): instr
            for mount, instr in self._instruments.items()
            if instr
        }

    @publish(command=cmds.pause)
    @requires_version(2, 0)
    def pause(self, msg: Optional[str] = None) -> None:
        """Pause execution of the protocol until it's resumed.

        A human can resume the protocol in the Opentrons App or on the touchscreen.

        This function returns immediately, but the next function call that
        is blocked by a paused robot (anything that involves moving) will
        not return until the protocol is resumed.

        :param str msg: An optional message to show in the run log entry for the pause step.
        """
        self._core.pause(msg=msg)

    @publish(command=cmds.resume)
    @requires_version(2, 0)
    def resume(self) -> None:
        """Resume the protocol after :py:meth:`pause`.

        .. deprecated:: 2.12
           The Python Protocol API supports no safe way for a protocol to resume itself.
           If you're looking for a way for your protocol to resume automatically
           after a period of time, use :py:meth:`delay`.
        """
        if self._api_version >= ENGINE_CORE_API_VERSION:
            raise APIVersionError(
                "A Python Protocol cannot safely resume itself after a pause."
                " To wait automatically for a period of time, use ProtocolContext.delay()."
            )

        # TODO(mc, 2023-02-13): this assert should be enough for mypy
        # investigate if upgrading mypy allows the `cast` to be removed
        assert isinstance(self._core, LegacyProtocolCore)
        cast(LegacyProtocolCore, self._core).resume()

    @publish(command=cmds.comment)
    @requires_version(2, 0)
    def comment(self, msg: str) -> None:
        """
        Add a user-readable message to the run log.

        The message is visible anywhere you can view the run log, including the Opentrons App and the touchscreen on Flex.

        .. note::

            The value of the message is computed during protocol analysis,
            so ``comment()`` can't communicate real-time information during the
            actual protocol run.
        """
        self._core.comment(msg=msg)

    @publish(command=cmds.delay)
    @requires_version(2, 0)
    def delay(
        self,
        seconds: float = 0,
        minutes: float = 0,
        msg: Optional[str] = None,
    ) -> None:
        """Delay protocol execution for a specific amount of time.

        :param float seconds: The time to delay in seconds.
        :param float minutes: The time to delay in minutes.

        If both ``seconds`` and ``minutes`` are specified, they will be added together.
        """
        delay_time = seconds + minutes * 60
        self._core.delay(seconds=delay_time, msg=msg)

    @requires_version(2, 0)
    def home(self) -> None:
        """Home the movement system of the robot."""
        self._core.home()

    @property
    def location_cache(self) -> Optional[Location]:
        """The cache used by the robot to determine where it last was."""
        return self._core.get_last_location()

    @location_cache.setter
    def location_cache(self, loc: Optional[Location]) -> None:
        self._core.set_last_location(loc)

    @property
    @requires_version(2, 0)
    def deck(self) -> Deck:
        """An interface to provide information about what's currently loaded on the deck.
        This object is useful for determining if a slot on the deck is free.

        This object behaves like a dictionary whose keys are the :ref:`deck slot <deck-slots>` names.
        For instance, ``deck[1]``, ``deck["1"]``, and ``deck["D1"]``
        will all return the object loaded in the front-left slot.

        The value for each key depends on what is loaded in the slot:
          - A :py:obj:`~opentrons.protocol_api.Labware` if the slot contains a labware.
          - A module context if the slot contains a hardware module.
          - ``None`` if the slot doesn't contain anything.

        A module that occupies multiple slots is set as the value for all of the
        relevant slots. Currently, the only multiple-slot module is the Thermocycler.
        When loaded, the :py:class:`ThermocyclerContext` object is the value for
        ``deck`` keys ``"A1"`` and ``"B1"`` on Flex, and ``7``, ``8``, ``10``, and
        ``11`` on OT-2. In API version 2.13 and earlier, only slot 7 keyed to the
        Thermocycler object, and slots 8, 10, and 11 keyed to ``None``.

        Rather than filtering the objects in the deck map yourself,
        you can also use :py:attr:`loaded_labwares` to get a dict of labwares
        and :py:attr:`loaded_modules` to get a dict of modules.

        For :ref:`advanced-control` *only*, you can delete an element of the ``deck`` dict.
        This only works for deck slots that contain labware objects. For example, if slot
        1 contains a labware, ``del protocol.deck["1"]`` will free the slot so you can
        load another labware there.

        .. warning::
            Deleting labware from a deck slot does not pause the protocol. Subsequent
            commands continue immediately. If you need to physically move the labware to
            reflect the new deck state, add a :py:meth:`.pause` or use
            :py:meth:`.move_labware` instead.

        .. versionchanged:: 2.14
           Includes the Thermocycler in all of the slots it occupies.

        .. versionchanged:: 2.15
           ``del`` sets the corresponding labware's location to ``OFF_DECK``.
        """
        return self._deck

    @property
    @requires_version(2, 0)
    def fixed_trash(self) -> Union[Labware, TrashBin]:
        """The trash fixed to slot 12 of an OT-2's deck.

        In API version 2.15 and earlier, the fixed trash is a :py:class:`.Labware` object with one well. Access it like labware in your protocol. For example, ``protocol.fixed_trash["A1"]``.

        In API version 2.15 only, Flex protocols have a fixed trash in slot A3.

        In API version 2.16 and later, the fixed trash only exists in OT-2 protocols. It is a :py:class:`.TrashBin` object, which doesn't have any wells. Trying to access ``fixed_trash`` in a Flex protocol will raise an error. See :ref:`configure-trash-bin` for details on using the movable trash in Flex protocols.

        .. versionchanged:: 2.16
            Returns a ``TrashBin`` object.
        """
        if self._api_version >= APIVersion(2, 16):
            if self._core.robot_type == "OT-3 Standard":
                raise APIVersionError(
                    "Fixed Trash is not supported on Flex protocols in API Version 2.16 and above."
                )
            disposal_locations = self._core.get_disposal_locations()
            if len(disposal_locations) == 0:
                raise NoTrashDefinedError(
                    "No trash container has been defined in this protocol."
                )
            if isinstance(disposal_locations[0], TrashBin):
                return disposal_locations[0]

        fixed_trash = self._core_map.get(self._core.fixed_trash)
        if fixed_trash is None:
            raise NoTrashDefinedError(
                "No trash container has been defined in this protocol."
            )

        return fixed_trash

    def _load_fixed_trash(self) -> None:
        fixed_trash_core = self._core.fixed_trash
        if fixed_trash_core is not None:
            fixed_trash = Labware(
                core=fixed_trash_core,
                api_version=self._api_version,
                protocol_core=self._core,
                core_map=self._core_map,
            )
            self._core_map.add(fixed_trash_core, fixed_trash)

    @requires_version(2, 5)
    def set_rail_lights(self, on: bool) -> None:
        """
        Controls the robot's ambient lighting (rail lights).

        :param bool on: If ``True``, turn on the lights; otherwise, turn them off.
        """
        self._core.set_rail_lights(on=on)

    @requires_version(2, 14)
    def define_liquid(
        self, name: str, description: Optional[str], display_color: Optional[str]
    ) -> Liquid:
        """
        Define a liquid within a protocol.

        :param str name: A human-readable name for the liquid.
        :param str description: An optional description of the liquid.
        :param str display_color: An optional hex color code, with hash included, to represent the specified liquid. Standard three-value, four-value, six-value, and eight-value syntax are all acceptable.

        :return: A :py:class:`~opentrons.protocol_api.Liquid` object representing the specified liquid.
        """
        return self._core.define_liquid(
            name=name,
            description=description,
            display_color=display_color,
        )

    @property
    @requires_version(2, 5)
    def rail_lights_on(self) -> bool:
        """Returns ``True`` if the robot's ambient lighting is on."""
        return self._core.get_rail_lights_on()

    @property
    @requires_version(2, 5)
    def door_closed(self) -> bool:
        """Returns ``True`` if the front door of the robot is closed."""
        return self._core.door_closed()


def _create_module_context(
    module_core: Union[ModuleCore, NonConnectedModuleCore],
    protocol_core: ProtocolCore,
    core_map: LoadedCoreMap,
    api_version: APIVersion,
    broker: LegacyBroker,
) -> ModuleTypes:
    module_cls: Optional[Type[ModuleTypes]] = None
    if isinstance(module_core, AbstractTemperatureModuleCore):
        module_cls = TemperatureModuleContext
    elif isinstance(module_core, AbstractMagneticModuleCore):
        module_cls = MagneticModuleContext
    elif isinstance(module_core, AbstractThermocyclerCore):
        module_cls = ThermocyclerContext
    elif isinstance(module_core, AbstractHeaterShakerCore):
        module_cls = HeaterShakerContext
    elif isinstance(module_core, AbstractMagneticBlockCore):
        module_cls = MagneticBlockContext
    else:
        assert False, "Unsupported module type"

    return module_cls(
        core=module_core,
        protocol_core=protocol_core,
        core_map=core_map,
        api_version=api_version,
        broker=broker,
    )



from typing import Optional, Dict

from opentrons.util.async_helpers import ensure_yield
from opentrons.drivers.types import Temperature
from opentrons.drivers.temp_deck.abstract import AbstractTempDeckDriver

TEMP_DECK_MODELS = {
    "temperatureModuleV1": "temp_deck_v1.1",
    "temperatureModuleV2": "temp_deck_v20",
}


class SimulatingDriver(AbstractTempDeckDriver):
    def __init__(self, sim_model: Optional[str] = None):
        self._temp = Temperature(target=None, current=0)
        self._port: Optional[str] = None
        self._model = TEMP_DECK_MODELS[sim_model] if sim_model else "temp_deck_v1.1"

    @ensure_yield
    async def set_temperature(self, celsius: float) -> None:
        self._temp.target = celsius
        self._temp.current = self._temp.target

    @ensure_yield
    async def get_temperature(self) -> Temperature:
        return self._temp

    @ensure_yield
    async def deactivate(self) -> None:
        self._temp = Temperature(target=None, current=23)

    @ensure_yield
    async def connect(self) -> None:
        pass

    @ensure_yield
    async def is_connected(self) -> bool:
        return True

    @ensure_yield
    async def disconnect(self) -> None:
        pass

    @ensure_yield
    async def enter_programming_mode(self) -> None:
        pass

    @ensure_yield
    async def get_device_info(self) -> Dict[str, str]:
        return {
            "serial": "dummySerialTD",
            "model": self._model,
            "version": "dummyVersionTD",
        }



"""Temperature Module protocol commands."""

from .set_target_temperature import (
    SetTargetTemperature,
    SetTargetTemperatureCreate,
    SetTargetTemperatureParams,
    SetTargetTemperatureResult,
    SetTargetTemperatureCommandType,
)

from .wait_for_temperature import (
    WaitForTemperature,
    WaitForTemperatureCreate,
    WaitForTemperatureParams,
    WaitForTemperatureResult,
    WaitForTemperatureCommandType,
)

from .deactivate import (
    DeactivateTemperature,
    DeactivateTemperatureCreate,
    DeactivateTemperatureParams,
    DeactivateTemperatureResult,
    DeactivateTemperatureCommandType,
)

__all__ = [
    # temperatureModule/setTargetTemperature
    "SetTargetTemperature",
    "SetTargetTemperatureCreate",
    "SetTargetTemperatureParams",
    "SetTargetTemperatureResult",
    "SetTargetTemperatureCommandType",
    # temperatureModule/waitForTemperature
    "WaitForTemperature",
    "WaitForTemperatureCreate",
    "WaitForTemperatureParams",
    "WaitForTemperatureResult",
    "WaitForTemperatureCommandType",
    # temperatureModule/deactivateTemperature
    "DeactivateTemperature",
    "DeactivateTemperatureCreate",
    "DeactivateTemperatureParams",
    "DeactivateTemperatureResult",
    "DeactivateTemperatureCommandType",
]



from typing import Optional, Dict

from opentrons.util.async_helpers import ensure_yield
from opentrons.drivers.thermocycler.abstract import AbstractThermocyclerDriver
from opentrons.drivers.types import Temperature, PlateTemperature, ThermocyclerLidStatus
from opentrons.hardware_control.modules.types import ThermocyclerModuleModel
from opentrons.drivers.asyncio.communication.errors import ErrorResponse


class SimulatingDriver(AbstractThermocyclerDriver):
    DEFAULT_TEMP = 23

    def __init__(self, model: Optional[str] = None) -> None:
        self._ramp_rate: Optional[float] = None
        self._lid_status = ThermocyclerLidStatus.OPEN
        self._lid_temperature = Temperature(current=self.DEFAULT_TEMP, target=None)
        self._plate_temperature = PlateTemperature(
            current=self.DEFAULT_TEMP, target=None, hold=None
        )
        self._model = model if model else "thermocyclerModuleV1"

    def model(self) -> str:
        return self._model

    @ensure_yield
    async def connect(self) -> None:
        pass

    @ensure_yield
    async def disconnect(self) -> None:
        pass

    @ensure_yield
    async def is_connected(self) -> bool:
        return True

    @ensure_yield
    async def open_lid(self) -> None:
        self._lid_status = ThermocyclerLidStatus.OPEN

    @ensure_yield
    async def close_lid(self) -> None:
        self._lid_status = ThermocyclerLidStatus.CLOSED

    @ensure_yield
    async def lift_plate(self) -> None:
        if self._model == ThermocyclerModuleModel.THERMOCYCLER_V1.value:
            raise NotImplementedError()
        if self._lid_status != ThermocyclerLidStatus.OPEN:
            raise ErrorResponse(port="sim_port", response="Lid is not open")
        self._lid_status = ThermocyclerLidStatus.OPEN

    @ensure_yield
    async def get_lid_status(self) -> ThermocyclerLidStatus:
        return self._lid_status

    @ensure_yield
    async def get_lid_temperature(self) -> Temperature:
        return self._lid_temperature

    @ensure_yield
    async def set_plate_temperature(
        self,
        temp: float,
        hold_time: Optional[float] = None,
        volume: Optional[float] = None,
    ) -> None:
        self._plate_temperature.target = temp
        self._plate_temperature.current = temp
        self._plate_temperature.hold = 0

    @ensure_yield
    async def get_plate_temperature(self) -> PlateTemperature:
        return self._plate_temperature

    @ensure_yield
    async def set_ramp_rate(self, ramp_rate: float) -> None:
        self._ramp_rate = ramp_rate

    @ensure_yield
    async def set_lid_temperature(self, temp: float) -> None:
        """Set the lid temperature in deg Celsius"""
        self._lid_temperature.target = temp
        self._lid_temperature.current = temp

    @ensure_yield
    async def deactivate_lid(self) -> None:
        self._lid_temperature.target = None
        self._lid_temperature.current = self.DEFAULT_TEMP

    @ensure_yield
    async def deactivate_block(self) -> None:
        self._plate_temperature.target = None
        self._plate_temperature.current = self.DEFAULT_TEMP
        self._plate_temperature.hold = None
        self._ramp_rate = None

    @ensure_yield
    async def deactivate_all(self) -> None:
        await self.deactivate_lid()
        await self.deactivate_block()

    @ensure_yield
    async def get_device_info(self) -> Dict[str, str]:
        return {
            "serial": "dummySerialTC",
            "model": "dummyModelTC",
            "version": "dummyVersionTC",
        }

    @ensure_yield
    async def enter_programming_mode(self) -> None:
        pass

    @ensure_yield
    async def jog_lid(self, angle: float) -> None:
        if self._model == ThermocyclerModuleModel.THERMOCYCLER_V1.value:
            raise NotImplementedError()
        self._lid_status = (
            ThermocyclerLidStatus.IN_BETWEEN
            if angle < 0
            else ThermocyclerLidStatus.OPEN
        )



"""Input file reading."""

import os
import pathlib
import platform
from dataclasses import dataclass
from typing import List, Optional, Sequence, Union

import anyio

from .input_file import AbstractInputFile
from .protocol_files_invalid_error import ProtocolFilesInvalidError


@dataclass(frozen=True)
class BufferedFile:
    """A file that has been read into memory."""

    name: str
    contents: bytes
    path: Optional[pathlib.Path]


class FileReadError(ProtocolFilesInvalidError):
    """An error raised if input files cannot be read."""


class FileReaderWriter:
    """Input file reader/writer interface."""

    @staticmethod
    async def read(
        files: Sequence[Union[AbstractInputFile, pathlib.Path]]
    ) -> List[BufferedFile]:
        """Read a set of input files into memory."""
        return [await _read_file(input_file=file) for file in files]

    @staticmethod
    async def write(directory: pathlib.Path, files: Sequence[BufferedFile]) -> None:
        """Write a set of previously buffered files to disk."""
        await anyio.Path(directory).mkdir(parents=True, exist_ok=True)

        for file in files:
            path = directory / file.name
            await _write_and_fsync_file(path=path, contents=file.contents)

        await _fsync_directory(path=directory)


async def _read_file(
    input_file: Union[AbstractInputFile, pathlib.Path]
) -> BufferedFile:
    if isinstance(input_file, pathlib.Path):
        path: Optional[pathlib.Path] = input_file
        filename = input_file.name
        contents = await anyio.Path(input_file).read_bytes()
    elif not input_file.filename:
        raise FileReadError("File was missing a name")
    else:
        path = None
        filename = input_file.filename
        async with anyio.wrap_file(input_file.file) as f:
            contents = await f.read()

    return BufferedFile(
        name=filename,
        contents=contents,
        path=path,
    )


async def _write_and_fsync_file(path: pathlib.Path, contents: bytes) -> None:
    async with await anyio.open_file(path, "wb") as file:
        await file.write(contents)
        await file.flush()
        await anyio.to_thread.run_sync(os.fsync, file.wrapped.fileno())


async def _fsync_directory(path: pathlib.Path) -> None:
    def _fsync_directory_sync() -> None:
        # We can't fsync directories on Windows because doing os.open() on them raises
        # PermissionError. https://stackoverflow.com/questions/21785127
        if platform.system() != "Windows":
            fd = os.open(path, os.O_RDONLY)
            try:
                os.fsync(fd)
            finally:
                os.close(fd)

    await anyio.to_thread.run_sync(_fsync_directory_sync)



"""ProtocolEngine class definition."""
from contextlib import AsyncExitStack
from logging import getLogger
from typing import Dict, Optional, Union

from opentrons.protocols.models import LabwareDefinition
from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.modules import AbstractModule as HardwareModuleAPI
from opentrons.hardware_control.types import PauseType as HardwarePauseType
from opentrons_shared_data.errors import (
    ErrorCodes,
    EnumeratedError,
)

from .errors import ProtocolCommandFailedError, ErrorOccurrence
from .errors.exceptions import EStopActivatedError
from . import commands, slot_standardization
from .resources import ModelUtils, ModuleDataProvider
from .types import (
    LabwareOffset,
    LabwareOffsetCreate,
    LabwareUri,
    ModuleModel,
    Liquid,
    HexColor,
    PostRunHardwareState,
    DeckConfigurationType,
    AddressableAreaLocation,
)
from .execution import (
    QueueWorker,
    create_queue_worker,
    DoorWatcher,
    HardwareStopper,
)
from .state import StateStore, StateView
from .plugins import AbstractPlugin, PluginStarter
from .actions import (
    ActionDispatcher,
    PlayAction,
    PauseAction,
    PauseSource,
    StopAction,
    FinishAction,
    FinishErrorDetails,
    QueueCommandAction,
    AddLabwareOffsetAction,
    AddLabwareDefinitionAction,
    AddLiquidAction,
    AddAddressableAreaAction,
    AddModuleAction,
    HardwareStoppedAction,
    ResetTipsAction,
    SetPipetteMovementSpeedAction,
    FailCommandAction,
)


_log = getLogger(__name__)


class ProtocolEngine:
    """Main ProtocolEngine class.

    A ProtocolEngine instance holds the state of a protocol as it executes,
    and manages calls to a command executor that actually implements the logic
    of the commands themselves.

    Lifetime:
        Instances are single-use. Each instance is associated with a single protocol,
        or a a single chain of robot control such as Labware Position Check.

    Concurrency:
        Instances live in `asyncio` event loops. Each instance must be constructed inside an
        event loop, and then must be interacted with exclusively through that
        event loop's thread--even for regular non-`async` methods, like `.pause()`.
        (This is because there are background async tasks that monitor state changes using
        primitives that aren't thread-safe. See ChangeNotifier.)
    """

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        state_store: StateStore,
        action_dispatcher: Optional[ActionDispatcher] = None,
        plugin_starter: Optional[PluginStarter] = None,
        queue_worker: Optional[QueueWorker] = None,
        model_utils: Optional[ModelUtils] = None,
        hardware_stopper: Optional[HardwareStopper] = None,
        door_watcher: Optional[DoorWatcher] = None,
        module_data_provider: Optional[ModuleDataProvider] = None,
    ) -> None:
        """Initialize a ProtocolEngine instance.

        Must be called while an event loop is active.

        This constructor does not inject provider implementations.
        Prefer the `create_protocol_engine()` factory function.
        """
        self._hardware_api = hardware_api
        self._state_store = state_store
        self._model_utils = model_utils or ModelUtils()

        self._action_dispatcher = action_dispatcher or ActionDispatcher(
            sink=self._state_store
        )
        self._plugin_starter = plugin_starter or PluginStarter(
            state=self._state_store,
            action_dispatcher=self._action_dispatcher,
        )
        self._queue_worker = queue_worker or create_queue_worker(
            hardware_api=hardware_api,
            state_store=self._state_store,
            action_dispatcher=self._action_dispatcher,
        )
        self._hardware_stopper = hardware_stopper or HardwareStopper(
            hardware_api=hardware_api,
            state_store=state_store,
        )
        self._door_watcher = door_watcher or DoorWatcher(
            state_store=state_store,
            hardware_api=hardware_api,
            action_dispatcher=self._action_dispatcher,
        )
        self._module_data_provider = module_data_provider or ModuleDataProvider()

        self._queue_worker.start()
        self._door_watcher.start()

    @property
    def state_view(self) -> StateView:
        """Get an interface to retrieve calculated state values."""
        return self._state_store

    def add_plugin(self, plugin: AbstractPlugin) -> None:
        """Add a plugin to the engine to customize behavior."""
        self._plugin_starter.start(plugin)

    def play(self, deck_configuration: Optional[DeckConfigurationType] = None) -> None:
        """Start or resume executing commands in the queue."""
        requested_at = self._model_utils.get_timestamp()
        # TODO(mc, 2021-08-05): if starting, ensure plungers motors are
        # homed if necessary
        action = self._state_store.commands.validate_action_allowed(
            PlayAction(requested_at=requested_at, deck_configuration=deck_configuration)
        )
        self._action_dispatcher.dispatch(action)

        if self._state_store.commands.get_is_door_blocking():
            self._hardware_api.pause(HardwarePauseType.PAUSE)
        else:
            self._hardware_api.resume(HardwarePauseType.PAUSE)

    def pause(self) -> None:
        """Pause executing commands in the queue."""
        action = self._state_store.commands.validate_action_allowed(
            PauseAction(source=PauseSource.CLIENT)
        )
        self._action_dispatcher.dispatch(action)
        self._hardware_api.pause(HardwarePauseType.PAUSE)

    def add_command(self, request: commands.CommandCreate) -> commands.Command:
        """Add a command to the `ProtocolEngine`'s queue.

        Arguments:
            request: The command type and payload data used to construct
                the command in state.

        Returns:
            The full, newly queued command.

        Raises:
            SetupCommandNotAllowed: the request specified a setup command,
                but the engine was not idle or paused.
            RunStoppedError: the run has been stopped, so no new commands
                may be added.
        """
        request = slot_standardization.standardize_command(
            request, self.state_view.config.robot_type
        )

        command_id = self._model_utils.generate_id()
        request_hash = commands.hash_command_params(
            create=request,
            last_hash=self._state_store.commands.get_latest_command_hash(),
        )

        action = self.state_view.commands.validate_action_allowed(
            QueueCommandAction(
                request=request,
                request_hash=request_hash,
                command_id=command_id,
                created_at=self._model_utils.get_timestamp(),
            )
        )
        self._action_dispatcher.dispatch(action)
        return self._state_store.commands.get(command_id)

    async def wait_for_command(self, command_id: str) -> None:
        """Wait for a command to be completed.

        Will also return if the engine was stopped before it reached the command.
        """
        await self._state_store.wait_for(
            self._state_store.commands.get_command_is_final,
            command_id=command_id,
        )

    async def add_and_execute_command(
        self, request: commands.CommandCreate
    ) -> commands.Command:
        """Add a command to the queue and wait for it to complete.

        The engine must be started by calling `play` before the command will
        execute. You only need to call `play` once.

        Arguments:
            request: The command type and payload data used to construct
                the command in state.

        Returns:
            The command. If the command completed, it will be succeeded or failed.
            If the engine was stopped before it reached the command,
            the command will be queued.
        """
        command = self.add_command(request)
        await self.wait_for_command(command.id)
        return self._state_store.commands.get(command.id)

    def estop(self, maintenance_run: bool) -> None:
        """Signal to the engine that an estop event occurred.

        If there are any queued commands for the engine, they will be marked
        as failed due to the estop event. If there aren't any queued commands
        *and* this is a maintenance run (which has commands queued one-by-one),
        a series of actions will mark the engine as Stopped. In either case the
        queue worker will be deactivated; the primary difference is that the former
        case will expect the protocol runner to `finish()` the engine, whereas the
        maintenance run will be put into a state wherein the engine can be discarded.
        """
        if self._state_store.commands.get_is_stopped():
            return
        current_id = (
            self._state_store.commands.state.running_command_id
            or self._state_store.commands.state.queued_command_ids.head(None)
        )

        if current_id is not None:
            fail_action = FailCommandAction(
                command_id=current_id,
                error_id=self._model_utils.generate_id(),
                failed_at=self._model_utils.get_timestamp(),
                error=EStopActivatedError(message="Estop Activated"),
            )
            self._action_dispatcher.dispatch(fail_action)

            # In the case where the running command was a setup command - check if there
            # are any pending *run* commands and, if so, clear them all
            current_id = self._state_store.commands.state.queued_command_ids.head(None)
            if current_id is not None:
                fail_action = FailCommandAction(
                    command_id=current_id,
                    error_id=self._model_utils.generate_id(),
                    failed_at=self._model_utils.get_timestamp(),
                    error=EStopActivatedError(message="Estop Activated"),
                )
                self._action_dispatcher.dispatch(fail_action)
            self._queue_worker.cancel()
        elif maintenance_run:
            stop_action = self._state_store.commands.validate_action_allowed(
                StopAction(from_estop=True)
            )
            self._action_dispatcher.dispatch(stop_action)
            hardware_stop_action = HardwareStoppedAction(
                completed_at=self._model_utils.get_timestamp(),
                finish_error_details=FinishErrorDetails(
                    error=EStopActivatedError(message="Estop Activated"),
                    error_id=self._model_utils.generate_id(),
                    created_at=self._model_utils.get_timestamp(),
                ),
            )
            self._action_dispatcher.dispatch(hardware_stop_action)
            self._queue_worker.cancel()
        else:
            _log.info("estop pressed before protocol was started, taking no action.")

    async def stop(self) -> None:
        """Stop execution immediately, halting all motion and cancelling future commands.

        After an engine has been `stop`'ed, it cannot be restarted.

        After a `stop`, you must still call `finish` to give the engine a chance
        to clean up resources and propagate errors.
        """
        action = self._state_store.commands.validate_action_allowed(StopAction())
        self._action_dispatcher.dispatch(action)
        self._queue_worker.cancel()
        if self._hardware_api.is_movement_execution_taskified():
            # We 'taskify' hardware controller movement functions when running protocols
            # that are not backed by the engine. Such runs cannot be stopped by cancelling
            # the queue worker and hence need to be stopped via the execution manager.
            # `cancel_execution_and_running_tasks()` sets the execution manager in a CANCELLED state
            # and cancels the running tasks, which raises an error and gets us out of the
            # run function execution, just like `_queue_worker.cancel()` does for
            # engine-backed runs.
            await self._hardware_api.cancel_execution_and_running_tasks()

    async def wait_until_complete(self) -> None:
        """Wait until there are no more commands to execute.

        Raises:
            CommandExecutionFailedError: if any protocol command failed.
        """
        await self._state_store.wait_for(
            condition=self._state_store.commands.get_all_commands_final
        )

    async def finish(
        self,
        error: Optional[Exception] = None,
        drop_tips_after_run: bool = True,
        set_run_status: bool = True,
        post_run_hardware_state: PostRunHardwareState = PostRunHardwareState.HOME_AND_STAY_ENGAGED,
    ) -> None:
        """Gracefully finish using the ProtocolEngine, waiting for it to become idle.

        The engine will finish executing its current command (if any),
        and then shut down. After an engine has been `finished`'ed, it cannot
        be restarted.

        This method should not raise. If any exceptions happened during execution that were not
        properly caught by the CommandExecutor, or if any exceptions happen during this
        `finish()` call, they should be saved as `.state_view.get_summary().errors`.

        Arguments:
            error: An error that caused the stop, if applicable.
            drop_tips_after_run: Whether to drop tips as part of cleanup.
            set_run_status: Whether to calculate a `success` or `failure` run status.
                If `False`, will set status to `stopped`.
            post_run_hardware_state: The state in which to leave the gantry and motors in
                after the run is over.
        """
        if self._state_store.commands.state.stopped_by_estop:
            # This handles the case where the E-stop was pressed while we were *not* in the middle
            # of some hardware interaction that would raise it as an exception. For example, imagine
            # we were paused between two commands, or imagine we were executing a very long run of
            # comment commands.
            drop_tips_after_run = False
            post_run_hardware_state = PostRunHardwareState.DISENGAGE_IN_PLACE
            if error is None:
                error = EStopActivatedError(message="Estop was activated during a run")

        if error:
            # If the run had an error, check if that error indicates an E-stop.
            # This handles the case where the run was in the middle of some hardware control
            # method and the hardware controller raised an E-stop error from it.
            #
            # To do this, we need to scan all the way through the error tree.
            # By the time E-stop error has gotten to us, it may have been wrapped in other errors,
            # so we need to unwrap them to uncover the E-stop error's inner beauty.
            #
            # We don't use self._hardware_api.get_estop_state() because the E-stop may have been
            # released by the time we get here.
            if isinstance(error, EnumeratedError):
                if self._code_in_error_tree(
                    root_error=error, code=ErrorCodes.E_STOP_ACTIVATED
                ) or self._code_in_error_tree(
                    # Request from the hardware team for the v7.0 betas: to help in-house debugging
                    # of pipette overpressure events, leave the pipette where it was like we do
                    # for E-stops.
                    root_error=error,
                    code=ErrorCodes.PIPETTE_OVERPRESSURE,
                ):
                    drop_tips_after_run = False
                    post_run_hardware_state = PostRunHardwareState.DISENGAGE_IN_PLACE

            error_details: Optional[FinishErrorDetails] = FinishErrorDetails(
                error_id=self._model_utils.generate_id(),
                created_at=self._model_utils.get_timestamp(),
                error=error,
            )
        else:
            error_details = None

        self._action_dispatcher.dispatch(
            FinishAction(error_details=error_details, set_run_status=set_run_status)
        )

        # We have a lot of independent things to tear down. If any teardown fails, we want
        # to continue with the rest, to avoid leaking resources or leaving the engine with a broken
        # state. We use an AsyncExitStack to avoid a gigantic try/finally tree. Note that execution
        # order will be backwards because the stack is first-in-last-out.
        exit_stack = AsyncExitStack()
        exit_stack.push_async_callback(self._plugin_starter.stop)  # Last step.
        exit_stack.push_async_callback(
            # Cleanup after hardware halt and reset the hardware controller
            self._hardware_stopper.do_stop_and_recover,
            post_run_hardware_state=post_run_hardware_state,
            drop_tips_after_run=drop_tips_after_run,
        )
        exit_stack.callback(self._door_watcher.stop)

        disengage_before_stopping = (
            False
            if post_run_hardware_state == PostRunHardwareState.STAY_ENGAGED_IN_PLACE
            else True
        )
        # Halt any movements immediately
        exit_stack.push_async_callback(
            self._hardware_stopper.do_halt,
            disengage_before_stopping=disengage_before_stopping,
        )
        exit_stack.push_async_callback(self._queue_worker.join)  # First step.
        try:
            # If any teardown steps failed, this will raise something.
            await exit_stack.aclose()
        except Exception as hardware_stopped_exception:
            _log.exception("Exception during post-run finish steps.")
            finish_error_details: Optional[FinishErrorDetails] = FinishErrorDetails(
                error_id=self._model_utils.generate_id(),
                created_at=self._model_utils.get_timestamp(),
                error=hardware_stopped_exception,
            )
        else:
            finish_error_details = None

        self._action_dispatcher.dispatch(
            HardwareStoppedAction(
                completed_at=self._model_utils.get_timestamp(),
                finish_error_details=finish_error_details,
            )
        )

    def add_labware_offset(self, request: LabwareOffsetCreate) -> LabwareOffset:
        """Add a new labware offset and return it.

        The added offset will apply to subsequent `LoadLabwareCommand`s.

        To retrieve offsets later, see `.state_view.labware`.
        """
        request = slot_standardization.standardize_labware_offset(
            request, self.state_view.config.robot_type
        )

        labware_offset_id = self._model_utils.generate_id()
        created_at = self._model_utils.get_timestamp()
        self._action_dispatcher.dispatch(
            AddLabwareOffsetAction(
                labware_offset_id=labware_offset_id,
                created_at=created_at,
                request=request,
            )
        )
        return self.state_view.labware.get_labware_offset(
            labware_offset_id=labware_offset_id
        )

    def add_labware_definition(self, definition: LabwareDefinition) -> LabwareUri:
        """Add a labware definition to the state for subsequent labware loads."""
        self._action_dispatcher.dispatch(
            AddLabwareDefinitionAction(definition=definition)
        )
        return self._state_store.labware.get_uri_from_definition(definition)

    def add_liquid(
        self,
        name: str,
        color: Optional[HexColor],
        description: Optional[str],
        id: Optional[str] = None,
    ) -> Liquid:
        """Add a liquid to the state for subsequent liquid loads."""
        if id is None:
            id = self._model_utils.generate_id()

        liquid = Liquid(
            id=id,
            displayName=name,
            description=(description or ""),
            displayColor=color,
        )

        self._action_dispatcher.dispatch(AddLiquidAction(liquid=liquid))
        return liquid

    def add_addressable_area(self, addressable_area_name: str) -> None:
        """Add an addressable area to state."""
        area = AddressableAreaLocation(addressableAreaName=addressable_area_name)
        self._action_dispatcher.dispatch(
            AddAddressableAreaAction(addressable_area=area)
        )

    def reset_tips(self, labware_id: str) -> None:
        """Reset the tip state of a given labware."""
        # TODO(mm, 2023-03-10): Safely raise an error if the given labware isn't a
        # tip rack?
        self._action_dispatcher.dispatch(ResetTipsAction(labware_id=labware_id))

    # TODO(mm, 2022-11-10): This is a method on ProtocolEngine instead of a command
    # as a quick hack to support Python protocols. We should consider making this a
    # command, or adding speed parameters to existing commands.
    # https://opentrons.atlassian.net/browse/RCORE-373
    def set_pipette_movement_speed(
        self, pipette_id: str, speed: Optional[float]
    ) -> None:
        """Set the speed of a pipette's X/Y/Z movements. Does not affect plunger speed.

        None will use the hardware API's default.
        """
        self._action_dispatcher.dispatch(
            SetPipetteMovementSpeedAction(pipette_id=pipette_id, speed=speed)
        )

    async def use_attached_modules(
        self,
        modules_by_id: Dict[str, HardwareModuleAPI],
    ) -> None:
        """Load attached modules directly into state, without locations."""
        actions = [
            AddModuleAction(
                module_id=module_id,
                serial_number=mod.device_info["serial"],
                definition=self._module_data_provider.get_definition(
                    ModuleModel(mod.model())
                ),
                module_live_data=mod.live_data,
            )
            for module_id, mod in modules_by_id.items()
        ]

        for a in actions:
            self._action_dispatcher.dispatch(a)

    # TODO(tz, 7-12-23): move this to shared data when we dont relay on ErrorOccurrence
    @staticmethod
    def _code_in_error_tree(
        root_error: Union[EnumeratedError, ErrorOccurrence], code: ErrorCodes
    ) -> bool:
        if isinstance(root_error, ErrorOccurrence):
            # ErrorOccurrence is not the same as the enumerated error exceptions. Check the
            # code by a string value.
            if root_error.errorCode == code.value.code:
                return True
            return any(
                ProtocolEngine._code_in_error_tree(wrapped, code)
                for wrapped in root_error.wrappedErrors
            )

        # From here we have an exception, can just check the code + recurse to wrapped errors.
        if root_error.code == code:
            return True

        if (
            isinstance(root_error, ProtocolCommandFailedError)
            and root_error.original_error is not None
        ):
            # For this specific EnumeratedError child, we recurse on the original_error field
            # in favor of the general error.wrapping field.
            return ProtocolEngine._code_in_error_tree(root_error.original_error, code)

        if len(root_error.wrapping) == 0:
            return False
        return any(
            ProtocolEngine._code_in_error_tree(wrapped_error, code)
            for wrapped_error in root_error.wrapping
        )



"""Common pipetting command base models."""
from pydantic import BaseModel, Field
from typing import Optional

from ..types import WellLocation, DeckPoint


class PipetteIdMixin(BaseModel):
    """Mixin for command requests that take a pipette ID."""

    pipetteId: str = Field(
        ...,
        description="Identifier of pipette to use for liquid handling.",
    )


class AspirateVolumeMixin(BaseModel):
    """Mixin for the `volume` field of aspirate commands."""

    volume: float = Field(
        ...,
        description="The amount of liquid to aspirate, in µL."
        " Must not be greater than the remaining available amount, which depends on"
        " the pipette (see `loadPipette`), its configuration (see `configureForVolume`),"
        " the tip (see `pickUpTip`), and the amount you've aspirated so far."
        " There is some tolerance for floating point rounding errors.",
        ge=0,
    )


class DispenseVolumeMixin(BaseModel):
    """Mixin for the `volume` field of dispense commands."""

    volume: float = Field(
        ...,
        description="The amount of liquid to dispense, in µL."
        " Must not be greater than the currently aspirated volume."
        " There is some tolerance for floating point rounding errors.",
        ge=0,
    )


class FlowRateMixin(BaseModel):
    """Mixin for command requests that take a flow rate."""

    flowRate: float = Field(
        ..., description="Speed in µL/s configured for the pipette", gt=0
    )


class WellLocationMixin(BaseModel):
    """Mixin for command requests that take a location that's somewhere in a well."""

    labwareId: str = Field(
        ...,
        description="Identifier of labware to use.",
    )
    wellName: str = Field(
        ...,
        description="Name of well to use in labware.",
    )
    wellLocation: WellLocation = Field(
        default_factory=WellLocation,
        description="Relative well location at which to perform the operation",
    )


class MovementMixin(BaseModel):
    """Mixin for command requests that move a pipette."""

    minimumZHeight: Optional[float] = Field(
        None,
        description=(
            "Optional minimal Z margin in mm."
            " If this is larger than the API's default safe Z margin,"
            " it will make the arc higher. If it's smaller, it will have no effect."
        ),
    )

    forceDirect: bool = Field(
        False,
        description=(
            "If true, moving from one labware/well to another"
            " will not arc to the default safe z,"
            " but instead will move directly to the specified location."
            " This will also force the `minimumZHeight` param to be ignored."
            " A 'direct' movement is in X/Y/Z simultaneously."
        ),
    )

    speed: Optional[float] = Field(
        None,
        description=(
            "Override the travel speed in mm/s."
            " This controls the straight linear speed of motion."
        ),
    )


class BaseLiquidHandlingResult(BaseModel):
    """Base properties of a liquid handling result."""

    volume: float = Field(
        ...,
        description="Amount of liquid in uL handled in the operation.",
        ge=0,
    )


class DestinationPositionResult(BaseModel):
    """Mixin for command results that move a pipette."""

    position: DeckPoint = Field(
        DeckPoint(x=0, y=0, z=0),
        description=(
            "The (x,y,z) coordinates of the pipette's critical point in deck space"
            " after the move was completed."
        ),
    )



import logging
from logging.config import dictConfig
import sys
from typing import Any, Dict

from opentrons.config import CONFIG, ARCHITECTURE, SystemArchitecture


def _host_config(level_value: int) -> Dict[str, Any]:
    serial_log_filename = CONFIG["serial_log_file"]
    api_log_filename = CONFIG["api_log_file"]
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "basic": {
                "format": (
                    "%(asctime)s %(name)s %(levelname)s [Line %(lineno)s] %(message)s"
                )
            }
        },
        "handlers": {
            "debug": {
                "class": "logging.StreamHandler",
                "formatter": "basic",
                "level": level_value,
            },
            "serial": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "basic",
                "filename": serial_log_filename,
                "maxBytes": 5000000,
                "level": logging.DEBUG,
                "backupCount": 3,
            },
            "api": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "basic",
                "filename": api_log_filename,
                "maxBytes": 1000000,
                "level": logging.DEBUG,
                "backupCount": 5,
            },
        },
        "loggers": {
            "opentrons": {
                "handlers": ["debug", "api"],
                "level": level_value,
            },
            "opentrons.deck_calibration": {
                "handlers": ["debug", "api"],
                "level": level_value,
            },
            "opentrons.drivers.asyncio.communication.serial_connection": {
                "handlers": ["serial"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            "opentrons_hardware.drivers.can_bus.can_messenger": {
                "handlers": ["serial"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            "opentrons_hardware.drivers.binary_usb.bin_serial": {
                "handlers": ["serial"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            "__main__": {"handlers": ["api"], "level": level_value},
        },
    }


def _buildroot_config(level_value: int) -> Dict[str, Any]:
    # Import systemd.journald here since it is generally unavailble on non
    # linux systems and we probably don't want to use it on linux desktops
    # either
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "message_only": {"format": "%(message)s"},
        },
        "handlers": {
            "api": {
                "class": "systemd.journal.JournalHandler",
                "level": logging.DEBUG,
                "formatter": "message_only",
                "SYSLOG_IDENTIFIER": "opentrons-api",
            },
            "serial": {
                "class": "systemd.journal.JournalHandler",
                "level": logging.DEBUG,
                "formatter": "message_only",
                "SYSLOG_IDENTIFIER": "opentrons-api-serial",
            },
            "can_serial": {
                "class": "systemd.journal.JournalHandler",
                "level": logging.DEBUG,
                "formatter": "message_only",
                "SYSLOG_IDENTIFIER": "opentrons-api-serial-can",
            },
            "usbbin_serial": {
                "class": "systemd.journal.JournalHandler",
                "level": logging.DEBUG,
                "formatter": "message_only",
                "SYSLOG_IDENTIFIER": "opentrons-api-serial-usbbin",
            },
        },
        "loggers": {
            "opentrons.drivers.asyncio.communication.serial_connection": {
                "handlers": ["serial"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            "opentrons": {
                "handlers": ["api"],
                "level": level_value,
            },
            "opentrons_hardware": {
                "handlers": ["api"],
                "level": level_value,
            },
            "opentrons_hardware.drivers.can_bus.can_messenger": {
                "handlers": ["can_serial"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            "opentrons_hardware.drivers.binary_usb.bin_serial": {
                "handlers": ["usbbin_serial"],
                "level": logging.DEBUG,
                "propagate": False,
            },
            "__main__": {"handlers": ["api"], "level": level_value},
        },
    }


def _config(arch: SystemArchitecture, level_value: int) -> Dict[str, Any]:
    return {
        SystemArchitecture.YOCTO: _buildroot_config,
        SystemArchitecture.BUILDROOT: _buildroot_config,
        SystemArchitecture.HOST: _host_config,
    }[arch](level_value)


def log_init(level_name: str) -> None:
    """
    Function that sets log levels and format strings. Checks for the
    OT_API_LOG_LEVEL environment variable otherwise defaults to INFO
    """
    fallback_log_level = "INFO"
    ot_log_level = level_name.upper()
    if ot_log_level not in logging._nameToLevel:
        sys.stderr.write(
            f"OT Log Level {ot_log_level} not found. "
            f"Defaulting to {fallback_log_level}\n"
        )
        ot_log_level = fallback_log_level
    level_value = logging._nameToLevel[ot_log_level]
    logging_config = _config(ARCHITECTURE, level_value)
    dictConfig(logging_config)



"""
hardware_control: The sole authority for controlling the hardware of an OT2.

The hardware_control module presents a unified api for the lowest level of
hardware command that takes into account the robot as a whole. For instance,
it presents an API for moving a specific pipette mount (not a specific motor
or axis)  to a deck-absolute point (not a Smoothie-coordinate point).

This module is not for use outside the opentrons api module. Higher-level
functions are available elsewhere.
"""
from .adapters import SynchronousAdapter
from .api import API
from .pause_manager import PauseManager
from .backends import Controller, Simulator
from .types import CriticalPoint, ExecutionState, OT3Mount
from .constants import DROP_TIP_RELEASE_DISTANCE
from .thread_manager import ThreadManager
from .execution_manager import ExecutionManager
from .threaded_async_lock import ThreadedAsyncLock, ThreadedAsyncForbidden
from .protocols import HardwareControlInterface, FlexHardwareControlInterface
from .instruments import AbstractInstrument, Gripper
from typing import Union
from .ot3_calibration import OT3Transforms
from .robot_calibration import RobotCalibration
from opentrons.config.types import RobotConfig, OT3Config

from opentrons.types import Mount

# TODO (lc 12-05-2022) We should 1. figure out if we need
# to globally export a class that is strictly used in the hardware controller
# and 2. how to properly export an ot2 and ot3 pipette.
from .instruments.ot2.pipette import Pipette

OT2HardwareControlAPI = HardwareControlInterface[RobotCalibration, Mount, RobotConfig]
OT3HardwareControlAPI = FlexHardwareControlInterface[
    OT3Transforms, Union[Mount, OT3Mount], OT3Config
]
HardwareControlAPI = Union[OT2HardwareControlAPI, OT3HardwareControlAPI]

ThreadManagedHardware = ThreadManager[HardwareControlAPI]
SyncHardwareAPI = SynchronousAdapter[HardwareControlAPI]

__all__ = [
    "API",
    "AbstractInstrument",
    "Controller",
    "Simulator",
    "Pipette",
    "Gripper",
    "PauseManager",
    "SynchronousAdapter",
    "HardwareControlAPI",
    "CriticalPoint",
    "DROP_TIP_RELEASE_DISTANCE",
    "ThreadManager",
    "ExecutionManager",
    "ExecutionState",
    "ThreadedAsyncLock",
    "ThreadedAsyncForbidden",
    "ThreadManagedHardware",
    "SyncHardwareAPI",
    "OT2HardwareControlAPI",
    "OT3HardwareControlAPI",
]



from abc import ABC, abstractmethod
from typing import Dict

from opentrons.drivers.types import Temperature, RPM, HeaterShakerLabwareLatchStatus


class AbstractHeaterShakerDriver(ABC):
    @abstractmethod
    async def connect(self) -> None:
        """Connect to heater-shaker"""
        ...

    @abstractmethod
    async def disconnect(self) -> None:
        """Disconnect from heater-shaker"""
        ...

    @abstractmethod
    async def is_connected(self) -> bool:
        """Check connection to heater-shaker"""
        ...

    @abstractmethod
    async def open_labware_latch(self) -> None:
        """Send open-labware-latch command"""
        ...

    @abstractmethod
    async def close_labware_latch(self) -> None:
        """Send close-labware-latch command"""
        ...

    @abstractmethod
    async def set_temperature(self, temperature: float) -> None:
        """Send set-temperature command"""
        ...

    @abstractmethod
    async def get_temperature(self) -> Temperature:
        """Send get-temperature command"""
        ...

    @abstractmethod
    async def set_rpm(self, rpm: int) -> None:
        """Send set-rpm command"""
        ...

    @abstractmethod
    async def get_rpm(self) -> RPM:
        """Send get-rpm command"""
        ...

    @abstractmethod
    async def get_labware_latch_status(self) -> HeaterShakerLabwareLatchStatus:
        """Send get-labware-latch-status command"""
        ...

    @abstractmethod
    async def home(self) -> None:
        """Send deactivate shaker command"""
        ...

    @abstractmethod
    async def deactivate_heater(self) -> None:
        """Send deactivate heater command"""
        ...

    @abstractmethod
    async def get_device_info(self) -> Dict[str, str]:
        """Send get device info command"""
        ...

    @abstractmethod
    async def enter_programming_mode(self) -> None:
        """Reboot into programming mode"""
        ...



"""setRailLights command request, result, and implementation models."""
from __future__ import annotations
from pydantic import BaseModel, Field
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from ..execution import RailLightsHandler

SetRailLightsCommandType = Literal["setRailLights"]


class SetRailLightsParams(BaseModel):
    """Payload required to set the rail lights on or off."""

    on: bool = Field(
        ...,
        description="The field that determines if the light is turned off or on.",
    )


class SetRailLightsResult(BaseModel):
    """Result data from the execution of a setRailLights command."""

    pass


class SetRailLightsImplementation(
    AbstractCommandImpl[SetRailLightsParams, SetRailLightsResult]
):
    """setRailLights command implementation."""

    def __init__(self, rail_lights: RailLightsHandler, **kwargs: object) -> None:
        self._rail_lights = rail_lights

    async def execute(self, params: SetRailLightsParams) -> SetRailLightsResult:
        """Dispatch a set lights command setting the state of the rail lights."""
        await self._rail_lights.set_rail_lights(params.on)
        return SetRailLightsResult()


class SetRailLights(BaseCommand[SetRailLightsParams, SetRailLightsResult]):
    """setRailLights command model."""

    commandType: SetRailLightsCommandType = "setRailLights"
    params: SetRailLightsParams
    result: Optional[SetRailLightsResult]

    _ImplementationCls: Type[SetRailLightsImplementation] = SetRailLightsImplementation


class SetRailLightsCreate(BaseCommandCreate[SetRailLightsParams]):
    """setRailLights command request model."""

    commandType: SetRailLightsCommandType = "setRailLights"
    params: SetRailLightsParams

    _CommandCls: Type[SetRailLights] = SetRailLights



""" protocol_api: The user-facing API for OT2 protocols.

This package defines classes and functions for access through a protocol to
control the OT2.

"""
from opentrons.protocols.api_support.definitions import (
    MAX_SUPPORTED_VERSION,
    MIN_SUPPORTED_VERSION,
    MIN_SUPPORTED_VERSION_FOR_FLEX,
)

from .protocol_context import ProtocolContext
from .deck import Deck
from .instrument_context import InstrumentContext
from .labware import Labware, Well
from .module_contexts import (
    ModuleContext,
    ThermocyclerContext,
    MagneticModuleContext,
    TemperatureModuleContext,
    HeaterShakerContext,
    MagneticBlockContext,
)
from ._liquid import Liquid
from ._types import OFF_DECK
from ._trash_bin import TrashBin
from ._waste_chute import WasteChute
from ._nozzle_layout import (
    COLUMN,
    ALL,
)

from .create_protocol_context import (
    create_protocol_context,
    ProtocolEngineCoreRequiredError,
)

__all__ = [
    "MAX_SUPPORTED_VERSION",
    "MIN_SUPPORTED_VERSION",
    "MIN_SUPPORTED_VERSION_FOR_FLEX",
    "ProtocolContext",
    "Deck",
    "ModuleContext",
    "InstrumentContext",
    "TemperatureModuleContext",
    "MagneticModuleContext",
    "ThermocyclerContext",
    "HeaterShakerContext",
    "MagneticBlockContext",
    "Labware",
    "TrashBin",
    "WasteChute",
    "Well",
    "Liquid",
    "COLUMN",
    "ALL",
    "OFF_DECK",
    # For internal Opentrons use only:
    "create_protocol_context",
    "ProtocolEngineCoreRequiredError",
]



"""Protocol engine state module."""

from .state import State, StateStore, StateView
from .state_summary import StateSummary
from .config import Config
from .commands import CommandState, CommandView, CommandSlice, CurrentCommand
from .labware import LabwareState, LabwareView
from .pipettes import PipetteState, PipetteView, HardwarePipette
from .modules import ModuleState, ModuleView, HardwareModule
from .module_substates import (
    MagneticModuleId,
    MagneticModuleSubState,
    HeaterShakerModuleId,
    HeaterShakerModuleSubState,
    TemperatureModuleId,
    TemperatureModuleSubState,
    ThermocyclerModuleId,
    ThermocyclerModuleSubState,
    ModuleSubStateType,
)
from .geometry import GeometryView
from .motion import MotionView, PipetteLocationData

__all__ = [
    # top level state value and interfaces
    "State",
    "StateStore",
    "StateView",
    "StateSummary",
    # static engine configuration
    "Config",
    # command state and values
    "CommandState",
    "CommandView",
    "CommandSlice",
    "CurrentCommand",
    # labware state and values
    "LabwareState",
    "LabwareView",
    # pipette state and values
    "PipetteState",
    "PipetteView",
    "HardwarePipette",
    # module state and values
    "ModuleState",
    "ModuleView",
    "HardwareModule",
    "MagneticModuleId",
    "MagneticModuleSubState",
    "HeaterShakerModuleId",
    "HeaterShakerModuleSubState",
    "TemperatureModuleId",
    "TemperatureModuleSubState",
    "ThermocyclerModuleId",
    "ThermocyclerModuleSubState",
    "ModuleSubStateType",
    # computed geometry state
    "GeometryView",
    # computed motion state
    "MotionView",
    "PipetteLocationData",
]



"""Module data resource provider."""
from typing import Dict
from opentrons.hardware_control.modules.module_calibration import (
    load_all_module_calibrations,
)
from opentrons_shared_data.module import load_definition

from opentrons.types import DeckSlotName
from ..types import (
    ModuleModel,
    ModuleDefinition,
    ModuleOffsetVector,
    ModuleOffsetData,
    DeckSlotLocation,
)


class ModuleDataProvider:
    """Module data provider."""

    @staticmethod
    def get_definition(model: ModuleModel) -> ModuleDefinition:
        """Get the module definition."""
        data = load_definition(model_or_loadname=model.value, version="3")
        return ModuleDefinition.parse_obj(data)

    @staticmethod
    def load_module_calibrations() -> Dict[str, ModuleOffsetData]:
        """Load the module calibration offsets."""
        module_calibrations: Dict[str, ModuleOffsetData] = dict()
        calibration_data = load_all_module_calibrations()
        for calibration in calibration_data:
            # NOTE module_id is really the module serial number, change this
            module_calibrations[calibration.module_id] = ModuleOffsetData(
                moduleOffsetVector=ModuleOffsetVector(
                    x=calibration.offset.x,
                    y=calibration.offset.y,
                    z=calibration.offset.z,
                ),
                location=DeckSlotLocation(
                    slotName=DeckSlotName.from_primitive(calibration.slot),
                ),
            )
        return module_calibrations



"""
This module handles configuration management. It keeps track of where the
configuration data is found in the robot, and is able to search for data to
construct an index if an existing index is not found. All other modules that
use persistent configuration data should use this module to read and write it.

The settings file defined here is opentrons.json. This file should be located

- On the robot, in /data
- Not on the robot, either in
  - the directory from which the python importing this module was launched
  - ~/.opentrons for the current user (where it will be written if nothing is
    found)

The keys in opentrons.json are defined by the CONFIG_ELEMENTS tuple below.
The keys in the file are the name elements of the CONFIG_ELEMENTS. They can
also be specified via environment variables, the names of which are
OT_API_${UPPERCASED_NAME_ELEMENT}. For instance, to override the
robot_settings_file option from an environment variable, you would set the
OT_API_ROBOT_CONFIG_FILE variable.

This module's interface to the rest of the system are the IS_* attributes and
the CONFIG attribute.
"""
import enum
import os
import json
import logging
from pathlib import Path
import re
import shutil
import subprocess
import sys
from enum import Enum, auto
from typing import Dict, NamedTuple, Optional, cast

_CONFIG_FILENAME = "config.json"
_LEGACY_INDICES = (
    Path("/mnt") / "usbdrive" / "config" / "index.json",
    Path("/data") / "index.json",
)

log = logging.getLogger(__file__)

IS_WIN = sys.platform.startswith("win")
IS_OSX = sys.platform == "darwin"
IS_LINUX = sys.platform.startswith("linux")
IS_ROBOT = bool(
    IS_LINUX
    and (os.environ.get("RUNNING_ON_PI") or os.environ.get("RUNNING_ON_VERDIN"))
)
#: This is the correct thing to check to see if we’re running on a robot
IS_VIRTUAL = bool(os.environ.get("ENABLE_VIRTUAL_SMOOTHIE"))


class SystemArchitecture(Enum):
    HOST = auto()
    BUILDROOT = auto()
    YOCTO = auto()


ROBOT_FIRMWARE_DIR: Optional[Path] = None
#: The path to firmware files for modules

ARCHITECTURE: SystemArchitecture = SystemArchitecture.HOST
#: The system architecture running

JUPYTER_NOTEBOOK_ROOT_DIR: Optional[Path] = None
#: The path to the root dir for Jupyter

JUPYTER_NOTEBOOK_LABWARE_DIR: Optional[Path] = None
#: The path to labware installs for jupyter

OT_SYSTEM_VERSION = "0.0.0"
#: The semver string of the system


if IS_ROBOT:
    if "OT_SYSTEM_VERSION" in os.environ:
        OT_SYSTEM_VERSION = os.environ["OT_SYSTEM_VERSION"]
        ARCHITECTURE = SystemArchitecture.YOCTO
    else:
        try:
            with open("/etc/VERSION.json") as vj:
                contents = json.load(vj)
            OT_SYSTEM_VERSION = contents["buildroot_version"]
            ARCHITECTURE = SystemArchitecture.BUILDROOT
        except Exception:
            log.exception("Could not find version file in /etc/VERSION.json")
    ROBOT_FIRMWARE_DIR = Path("/usr/lib/firmware/")
    JUPYTER_NOTEBOOK_ROOT_DIR = Path("/var/lib/jupyter/notebooks/")
    JUPYTER_NOTEBOOK_LABWARE_DIR = JUPYTER_NOTEBOOK_ROOT_DIR / "labware"


def name() -> str:
    if IS_ROBOT and ARCHITECTURE in (
        SystemArchitecture.BUILDROOT,
        SystemArchitecture.YOCTO,
    ):
        # Read the name from the machine's pretty hostname, which is maintained
        # by update-server. This retrieval logic needs to be kept in sync with
        # update-server.

        # NOTE: This call to hostnamectl can fail momentarily if it runs
        # at the same time as a systemd-hostnamed restart.
        # update-server triggers such restarts regularly, any time the name changes.
        #
        # We let the exception propagate so the caller (probably an HTTP client
        # polling /health) can retry later.
        result = subprocess.check_output(["hostnamectl", "--pretty", "status"]).decode(
            "utf-8"
        )

        # Strip the trailing newline, since it's not part of the actual name value.
        # TODO(mm, 2022-07-18): When we upgrade to systemd 249, use
        # `hostnamectl --json` for CLI output that we can parse more robustly.
        assert len(result) >= 1 and result[-1] == "\n"
        return result[:-1]

    else:
        robot_name = "opentrons-dev"
        if "DEV_ROBOT_NAME" in os.environ.keys():
            robot_name = os.environ["DEV_ROBOT_NAME"]
        return robot_name


class ConfigElementType(enum.Enum):
    FILE = enum.auto()
    DIR = enum.auto()


class ConfigElement(NamedTuple):
    name: str
    display_name: str
    default: Path
    kind: ConfigElementType
    help: str


CONFIG_ELEMENTS = (
    ConfigElement(
        "labware_database_file",
        "API V1 Labware Database",
        Path("opentrons.db"),
        ConfigElementType.FILE,
        (
            "A SQLite database where labware definitions and offsets were stored."
            " No longer used as of v4."
        ),
    ),
    ConfigElement(
        "labware_calibration_offsets_dir_v2",
        "API V2 Calibration Offsets Directory",
        Path("labware") / "v2" / "offsets",
        ConfigElementType.DIR,
        "A location where labware offsets were stored. No longer used as of v5.",
    ),
    ConfigElement(
        "labware_user_definitions_dir_v2",
        "API V2 Custom Labware Directory",
        Path("labware") / "v2" / "custom_definitions",
        ConfigElementType.DIR,
        (
            "A location where custom labware definitions may be stored."
            " Usage not recommended."
        ),
    ),
    ConfigElement(
        "feature_flags_file",
        "Feature Flags",
        Path("feature_flags.json"),
        ConfigElementType.FILE,
        "The file storing the advanced feature flags.",
    ),
    ConfigElement(
        "robot_settings_file",
        "Robot Settings",
        Path("robot_settings.json"),
        ConfigElementType.FILE,
        "The file storing settings relevant to motion",
    ),
    ConfigElement(
        "deck_calibration_file",
        "Deck Calibration",
        Path("deck_calibration.json"),
        ConfigElementType.FILE,
        "The file storing the deck calibration. Superseded in v4 by robot_calibration_dir.",
    ),
    ConfigElement(
        "log_dir",
        "Log Directory",
        Path("logs"),
        ConfigElementType.FILE,
        "The location for saving log files",
    ),
    ConfigElement(
        "api_log_file",
        "API Log File",
        Path("logs") / "api.log",
        ConfigElementType.FILE,
        "The location of the file to save API logs to. If this is an"
        " absolute path, it will be used directly. If it is a "
        "relative path it will be relative to log_dir",
    ),
    ConfigElement(
        "serial_log_file",
        "Serial Log File",
        Path("logs") / "serial.log",
        ConfigElementType.FILE,
        "The location of the file to save serial logs to. If this is"
        " an absolute path, it will be used directly. If it is a "
        "relative path it will be relative to log_dir"
        "The location of the file to save serial logs to",
    ),
    # Unlike other config elements, the wifi keys dir is still in
    # /data/user_storage/opentrons_data because these paths are fed directly to
    # NetworkManager and stored in connections files there. To change this
    # directory, we would have to modify those connections files, presumably on
    # boot, which is a level of complexity that makes it worth having an
    # annoying path.
    ConfigElement(
        "wifi_keys_dir",
        "Wifi Keys Dir",
        Path("user_storage/opentrons_data/network_keys"),
        ConfigElementType.DIR,
        "The directory in which to save any key material for wifi"
        " auth. Not relevant outside of a robot.",
    ),
    ConfigElement(
        "hardware_controller_lockfile",
        "Hardware Controller Lockfile",
        Path("hardware.lock"),
        ConfigElementType.FILE,
        "The file to use for a hardware controller lockfile.",
    ),
    ConfigElement(
        "pipette_config_overrides_dir",
        "Pipette Config User Overrides",
        Path("pipettes"),
        ConfigElementType.DIR,
        "The dir where settings overrides for pipettes are stored",
    ),
    ConfigElement(
        "tip_length_calibration_dir",
        "Tip Length Calibration Directory",
        Path("tip_lengths"),
        ConfigElementType.DIR,
        "The dir where tip length calibration of each tiprack for "
        "each unique pipette is stored",
    ),
    ConfigElement(
        "robot_calibration_dir",
        "Robot Calibration Directory",
        Path("robot"),
        ConfigElementType.DIR,
        "The dir where robot calibration is stored",
    ),
    ConfigElement(
        "pipette_calibration_dir",
        "Pipette Calibration Directory",
        Path("robot") / "pipettes",
        ConfigElementType.DIR,
        "The dir where pipette calibration is stored",
    ),
    ConfigElement(
        "custom_tiprack_dir",
        "Custom Tiprack Directory",
        Path("tip_lengths") / "custom_tiprack_definitions",
        ConfigElementType.DIR,
        "The dir where custom tiprack definitions for tip length "
        "calibration are stored",
    ),
    ConfigElement(
        "gripper_calibration_dir",
        "Gripper Calibration Directory",
        Path("robot") / "gripper",
        ConfigElementType.DIR,
        "The dir where gripper calibration is stored",
    ),
    ConfigElement(
        "module_calibration_dir",
        "Module Calibration Directory",
        Path("robot") / "modules",
        ConfigElementType.DIR,
        "The dir where module calibration is stored",
    ),
)
#: The available configuration file elements to modify. All of these can be
#: changed by editing opentrons.json, where the keys are the name elements,
#: or by specifying as environment variables, where the keys are uppercase
#: versions of the name elements.
#: In addition to these flags, the OT_API_CONFIG_DIR env var (if present)
#: will change where the API looks for these settings by prepending it to the
#: normal search path.


def infer_config_base_dir() -> Path:
    """Return the directory to store data in.

    Defaults are ~/.opentrons if not on a pi; OT_API_CONFIG_DIR is
    respected here.

    When this module is imported, this function is called automatically
    and the result stored in :py:attr:`APP_DATA_DIR`.

    This directory may not exist when the module is imported. Even if it
    does exist, it may not contain data, or may require data to be moved
    to it.

    :return pathlib.Path: The path to the desired root settings dir.
    """
    if "OT_API_CONFIG_DIR" in os.environ:
        return Path(os.environ["OT_API_CONFIG_DIR"])
    elif IS_ROBOT:
        return Path("/data")
    else:
        search = (Path.cwd(), Path.home() / ".opentrons")
        for path in search:
            if (path / _CONFIG_FILENAME).exists():
                return path
        else:
            return search[-1]


def load_and_migrate() -> Dict[str, Path]:
    """Ensure the settings directory tree is properly configured.

    This function does most of its work on the actual robot. It will move
    all settings files from wherever they happen to be to the proper
    place. On non-robots, this mostly just loads. In addition, it writes
    a default config and makes sure all directories required exist (though
    the files in them may not).
    """
    if IS_ROBOT and ARCHITECTURE != SystemArchitecture.YOCTO:
        _migrate_robot()
    base = infer_config_base_dir()
    base.mkdir(parents=True, exist_ok=True)
    index = _load_with_overrides(base)
    return _ensure_paths_and_types(index)


def _load_with_overrides(base: Path) -> Dict[str, str]:
    """Load an config or write its defaults"""
    should_write = False
    overrides = _get_environ_overrides()
    try:
        with (base / _CONFIG_FILENAME).open() as file:
            index = json.load(file)
    except (OSError, json.JSONDecodeError):
        should_write = True
        index = generate_config_index(overrides)

    for key in CONFIG_ELEMENTS:
        if key.name not in index:
            if key.kind in (ConfigElementType.DIR, ConfigElementType.FILE):
                index[key.name] = base / key.default
            else:
                index[key.name] = key.default
            should_write = True

    if should_write:
        try:
            write_config(index, path=base)
        except Exception as e:
            sys.stderr.write(
                "Error writing config to {}: {}\nProceeding memory-only\n".format(
                    str(base), e
                )
            )
    index.update(overrides)
    return cast(Dict[str, str], index)


def _ensure_paths_and_types(index: Dict[str, str]) -> Dict[str, Path]:
    """Take the direct results of loading the config and make sure
    the filesystem reflects them.
    """
    configs_by_name = {ce.name: ce for ce in CONFIG_ELEMENTS}
    correct_types: Dict[str, Path] = {}
    for key, item in index.items():
        if key not in configs_by_name:  # old config, ignore
            continue
        if configs_by_name[key].kind == ConfigElementType.FILE:
            it = Path(item)
            it.parent.mkdir(parents=True, exist_ok=True)
            correct_types[key] = it
        elif configs_by_name[key].kind == ConfigElementType.DIR:
            it = Path(item)
            it.mkdir(parents=True, exist_ok=True)
            correct_types[key] = it
        else:
            raise RuntimeError(
                f"unhandled kind in ConfigElements: {key}: "
                f"{configs_by_name[key].kind}"
            )
    return correct_types


def _get_environ_overrides() -> Dict[str, str]:
    """Pull any overrides for the config elements from the environ and return
    a mapping from the names to the values (as strings). Config elements that
    are not overridden will not be in the mapping.
    """
    return {
        ce.name: os.environ["OT_API_" + ce.name.upper()]
        for ce in CONFIG_ELEMENTS
        if "OT_API_" + ce.name.upper() in os.environ
    }


def _legacy_index() -> Optional[Dict[str, str]]:
    """Try and load an index file from the various places it might exist.

    If the legacy file cannot be found or cannot be parsed, return None.

    This method should only be called on a robot.
    """
    for index in _LEGACY_INDICES:
        if index.exists():
            try:
                with open(index) as file:
                    return cast(Dict[str, str], json.load(file))
            except (OSError, json.JSONDecodeError):
                return None
    return None


def _erase_old_indices() -> None:
    """Remove old index files so they don't pollute future loads.

    This method should only be called on a robot.
    """
    for index in _LEGACY_INDICES:
        if index.exists():
            index.unlink()


def _find_most_recent_backup(normal_path: Optional[str]) -> Optional[str]:
    """Find the most recent old settings to migrate.

    The input is the path to an unqualified settings file - e.g.
    /mnt/usbdrive/config/robotSettings.json

    This will return
    - None if the input is None (to support chaining from dict.get())
    - The input if it exists, or
    - The file named normal_path-TIMESTAMP.json with the highest timestamp
      if one can be found, or
    - None
    """
    if normal_path is None:
        return None

    if os.path.exists(normal_path):
        return normal_path

    dirname, basename = os.path.split(normal_path)
    root, ext = os.path.splitext(basename)
    backups = [
        fi for fi in os.listdir(dirname) if fi.startswith(root) and fi.endswith(ext)
    ]
    ts_re = re.compile(r".*-([0-9]+)" + ext + "$")

    def ts_compare(filename: str) -> int:
        match = ts_re.match(filename)
        if not match:
            return -1
        else:
            return int(match.group(1))

    backups_sorted = sorted(backups, key=ts_compare)
    if not backups_sorted:
        return None
    return os.path.join(dirname, backups_sorted[-1])


def _do_migrate(index: Dict[str, str]) -> None:
    base = infer_config_base_dir()
    new_index = generate_config_index(_get_environ_overrides(), base)
    moves = (
        (
            "/data/user_storage/opentrons_data/opentrons.db",
            new_index["labware_database_file"],
        ),
        (
            _find_most_recent_backup(index.get("robotSettingsFile")),
            new_index["robot_settings_file"],
        ),
        (index.get("deckCalibrationFile"), new_index["deck_calibration_file"]),
        (index.get("featureFlagFile"), new_index["feature_flags_file"]),
    )
    sys.stdout.write(f"config migration: new base {base}\n")
    for old, new in moves:
        if not old:
            continue
        old_path = Path(old)
        new_path = Path(new)
        if old_path.exists() and not old_path.is_symlink():
            sys.stdout.write(f"config migration: {old}->{new}\n")
            if new_path.is_dir():
                shutil.rmtree(new_path)
            shutil.move(str(old_path), str(new_path))
        else:
            sys.stdout.write(f"config migration: not moving {old}:")
            sys.stdout.write(f" exists={old_path.exists()}")
            sys.stdout.write(f" symlink={old_path.is_symlink()}\n")

    write_config(new_index, base)


def _migrate_robot() -> None:
    old_index = _legacy_index()
    if old_index:
        _do_migrate(old_index)
        _erase_old_indices()


def generate_config_index(
    defaults: Dict[str, str], base_dir: Optional[Path] = None
) -> Dict[str, Path]:
    """
    Determines where existing info can be found in the system, and creates a
    corresponding data dict that can be written to index.json in the
    baseDataDir.

    The information in the files defined by the config index is information
    required by the API itself and nothing else - labware definitions, feature
    flags, robot configurations. It does not include configuration files that
    relate to the rest of the system, such as network description file
    definitions.

    :param defaults: A dict of defaults to write, useful for specifying part
                     (but not all) of the index succinctly. This is used both
                     when loading a configuration file from disk and when
                     generating a new one.
    :param base_dir: If specified, a base path used if this function has to
                     generate defaults. If not specified, falls back to
                     :py:attr:`CONFIG_BASE_DIR`
    :returns: The config object
    """
    base = Path(base_dir) if base_dir else infer_config_base_dir()

    def parse_or_default(ce: ConfigElement, val: Optional[str]) -> Path:
        if not val:
            return base / ce.default
        else:
            return Path(val)

    return {
        ce.name: parse_or_default(ce, defaults.get(ce.name)) for ce in CONFIG_ELEMENTS
    }


def write_config(config_data: Dict[str, Path], path: Optional[Path] = None) -> None:
    """Save the config file.

    :param config_data: The index to save
    :param base_dir: The place to save the file. If ``None``,
                     :py:meth:`infer_config_base_dir()` will be used

    Only keys that are in the config elements will be saved.
    """
    path = Path(path) if path else infer_config_base_dir()
    valid_names = [ce.name for ce in CONFIG_ELEMENTS]
    try:
        os.makedirs(path, exist_ok=True)
        with (path / _CONFIG_FILENAME).open("w") as base_f:
            json.dump(
                {k: str(v) for k, v in config_data.items() if k in valid_names},
                base_f,
                indent=2,
            )
    except OSError as e:
        sys.stderr.write(
            "Config index write to {} failed: {}\n".format(path / _CONFIG_FILENAME, e)
        )


def reload() -> None:
    global CONFIG
    CONFIG.clear()
    CONFIG.update(load_and_migrate())


def get_opentrons_path(path_name: str) -> Path:
    # Helper function to look-up the path
    # to specific configuration files for
    # the Opentrons system
    global CONFIG
    return CONFIG[path_name]


CONFIG = load_and_migrate()
#: The currently loaded config. This should not change for the lifetime
#: of the program. This is a dict much like os.environ() where the keys
#: are config element names


def get_tip_length_cal_path() -> Path:
    return get_opentrons_path("tip_length_calibration_dir")


def get_custom_tiprack_def_path() -> Path:
    return get_opentrons_path("custom_tiprack_dir")



from typing import Optional


class SmoothieError(Exception):
    def __init__(
        self, ret_code: Optional[str] = None, command: Optional[str] = None
    ) -> None:
        self.ret_code = ret_code or ""
        self.command = command
        super().__init__()

    def __repr__(self) -> str:
        return f"<SmoothieError: {self.ret_code} from {self.command}>"

    def __str__(self) -> str:
        return f"SmoothieError: {self.command} returned {self.ret_code}"


class SmoothieAlarm(Exception):
    def __init__(
        self, ret_code: Optional[str] = None, command: Optional[str] = None
    ) -> None:
        self.ret_code = ret_code
        self.command = command
        super().__init__()

    def __repr__(self) -> str:
        return f"<SmoothieAlarm: {self.ret_code} from {self.command}>"

    def __str__(self) -> str:
        return f"SmoothieAlarm: {self.command} returned {self.ret_code}"


class TipProbeError(SmoothieAlarm):
    def __init__(
        self, ret_code: Optional[str] = None, command: Optional[str] = None
    ) -> None:
        self.ret_code = ret_code
        self.command = command
        super().__init__(ret_code, command)

    def __repr__(self) -> str:
        return f"<TipProbeError: {self.ret_code} from {self.command}"

    def __str__(self) -> str:
        return (
            "Tip probe could not complete: the switch was never touched. "
            "This may be because there is no tip on the pipette."
        )



"""Implementation, request models, and response models for the load module command."""
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type
from typing_extensions import Literal
from pydantic import BaseModel, Field

from .command import AbstractCommandImpl, BaseCommand, BaseCommandCreate
from ..types import DeckSlotLocation, ModuleModel, ModuleDefinition

if TYPE_CHECKING:
    from ..state import StateView
    from ..execution import EquipmentHandler


LoadModuleCommandType = Literal["loadModule"]


class LoadModuleParams(BaseModel):
    """Payload required to load a module."""

    model: ModuleModel = Field(
        ...,
        description=(
            "The model name of the module to load."
            "\n\n"
            "Protocol Engine will look for a connected module that either"
            " exactly matches this one, or is compatible."
            "\n\n"
            " For example, if you request a `temperatureModuleV1` here,"
            " Protocol Engine might load a `temperatureModuleV1` or a `temperatureModuleV2`."
            "\n\n"
            " The model that it finds connected will be available through `result.model`."
        ),
    )

    # Note: Our assumption here that a module's position can be boiled down to a
    # single deck slot precludes loading a Thermocycler in its special "shifted slightly
    # to the left" position. This is okay for now because neither the Python Protocol
    # API nor Protocol Designer attempt to support it, either.
    location: DeckSlotLocation = Field(
        ...,
        description=(
            "The location into which this module should be loaded."
            "\n\n"
            "For the Thermocycler Module, which occupies multiple deck slots,"
            " this should be the front-most occupied slot (normally slot 7)."
        ),
    )

    moduleId: Optional[str] = Field(
        None,
        description=(
            "An optional ID to assign to this module."
            " If None, an ID will be generated."
        ),
    )


class LoadModuleResult(BaseModel):
    """The results of loading a module."""

    moduleId: str = Field(
        description="An ID to reference this module in subsequent commands."
    )

    # TODO(mm, 2023-04-13): Remove this field. Jira RSS-221.
    definition: ModuleDefinition = Field(
        deprecated=True,
        description=(
            "The definition of the connected module."
            " This field is an implementation detail. We might change or remove it without warning."
            " Do not access it or rely on it being present."
        ),
    )

    model: ModuleModel = Field(
        ...,
        description=(
            "The hardware model of the connected module."
            " This can be different from the exact model that this command requested."
            " See `params.model`."
            "\n\n"
            "This field is only meaningful in the run's actual execution,"
            " not in the protocol's analysis."
            " In analysis, it will be an arbitrary placeholder."
        ),
    )

    serialNumber: Optional[str] = Field(
        None,
        description="Hardware serial number of the connected module. "
        "Will be `None` if a module is not electrically connected to the robot (like the Magnetic Block).",
    )


class LoadModuleImplementation(AbstractCommandImpl[LoadModuleParams, LoadModuleResult]):
    """The implementation of the load module command."""

    def __init__(
        self, equipment: EquipmentHandler, state_view: StateView, **kwargs: object
    ) -> None:
        self._equipment = equipment
        self._state_view = state_view

    async def execute(self, params: LoadModuleParams) -> LoadModuleResult:
        """Check that the requested module is attached and assign its identifier."""
        self._state_view.addressable_areas.raise_if_area_not_in_deck_configuration(
            params.location.slotName.id
        )

        verified_location = self._state_view.geometry.ensure_location_not_occupied(
            params.location
        )

        if params.model == ModuleModel.MAGNETIC_BLOCK_V1:
            loaded_module = await self._equipment.load_magnetic_block(
                model=params.model,
                location=verified_location,
                module_id=params.moduleId,
            )
        else:
            loaded_module = await self._equipment.load_module(
                model=params.model,
                location=verified_location,
                module_id=params.moduleId,
            )

        return LoadModuleResult(
            moduleId=loaded_module.module_id,
            serialNumber=loaded_module.serial_number,
            model=loaded_module.definition.model,
            definition=loaded_module.definition,
        )


class LoadModule(BaseCommand[LoadModuleParams, LoadModuleResult]):
    """The model for a load module command."""

    commandType: LoadModuleCommandType = "loadModule"
    params: LoadModuleParams
    result: Optional[LoadModuleResult]

    _ImplementationCls: Type[LoadModuleImplementation] = LoadModuleImplementation


class LoadModuleCreate(BaseCommandCreate[LoadModuleParams]):
    """The model for a creation request for a load module command."""

    commandType: LoadModuleCommandType = "loadModule"
    params: LoadModuleParams

    _CommandCls: Type[LoadModule] = LoadModule



"""Command models to close the Heater-Shaker Module's labware latch."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from typing_extensions import Literal, Type

from pydantic import BaseModel, Field

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate

if TYPE_CHECKING:
    from opentrons.protocol_engine.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler


CloseLabwareLatchCommandType = Literal["heaterShaker/closeLabwareLatch"]


class CloseLabwareLatchParams(BaseModel):
    """Input parameters to close a Heater-Shaker Module's labware latch."""

    moduleId: str = Field(..., description="Unique ID of the Heater-Shaker Module.")


class CloseLabwareLatchResult(BaseModel):
    """Result data from closing a Heater-Shaker's labware latch."""


class CloseLabwareLatchImpl(
    AbstractCommandImpl[CloseLabwareLatchParams, CloseLabwareLatchResult]
):
    """Execution implementation of a Heater-Shaker's close labware latch command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(self, params: CloseLabwareLatchParams) -> CloseLabwareLatchResult:
        """Close a Heater-Shaker's labware latch."""
        # Allow propagation of ModuleNotLoadedError and WrongModuleTypeError.
        hs_module_substate = self._state_view.modules.get_heater_shaker_module_substate(
            module_id=params.moduleId
        )

        # Allow propagation of ModuleNotAttachedError.
        hs_hardware_module = self._equipment.get_module_hardware_api(
            hs_module_substate.module_id
        )

        if hs_hardware_module is not None:
            await hs_hardware_module.close_labware_latch()

        return CloseLabwareLatchResult()


class CloseLabwareLatch(BaseCommand[CloseLabwareLatchParams, CloseLabwareLatchResult]):
    """A command to close a Heater-Shaker's latch."""

    commandType: CloseLabwareLatchCommandType = "heaterShaker/closeLabwareLatch"
    params: CloseLabwareLatchParams
    result: Optional[CloseLabwareLatchResult]

    _ImplementationCls: Type[CloseLabwareLatchImpl] = CloseLabwareLatchImpl


class CloseLabwareLatchCreate(BaseCommandCreate[CloseLabwareLatchParams]):
    """A request to create a Heater-Shaker's close latch command."""

    commandType: CloseLabwareLatchCommandType = "heaterShaker/closeLabwareLatch"
    params: CloseLabwareLatchParams

    _CommandCls: Type[CloseLabwareLatch] = CloseLabwareLatch



"""Command side-effect execution logic container."""
import asyncio
from logging import getLogger
from typing import Optional

from opentrons.hardware_control import HardwareControlAPI

from opentrons_shared_data.errors.exceptions import (
    EStopActivatedError,
    EnumeratedError,
    PythonException,
)

from ..state import StateStore
from ..resources import ModelUtils
from ..commands import (
    CommandStatus,
    AbstractCommandImpl,
    CommandResult,
    CommandPrivateResult,
)
from ..actions import ActionDispatcher, UpdateCommandAction, FailCommandAction
from ..errors import RunStoppedError
from ..errors.exceptions import EStopActivatedError as PE_EStopActivatedError
from .equipment import EquipmentHandler
from .movement import MovementHandler
from .gantry_mover import GantryMover
from .labware_movement import LabwareMovementHandler
from .pipetting import PipettingHandler
from .tip_handler import TipHandler
from .run_control import RunControlHandler
from .rail_lights import RailLightsHandler
from .status_bar import StatusBarHandler


log = getLogger(__name__)


class CommandExecutor:
    """CommandExecutor container class.

    CommandExecutor manages various child handlers that define procedures to
    execute the side-effects of commands.
    """

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        state_store: StateStore,
        action_dispatcher: ActionDispatcher,
        equipment: EquipmentHandler,
        movement: MovementHandler,
        gantry_mover: GantryMover,
        labware_movement: LabwareMovementHandler,
        pipetting: PipettingHandler,
        tip_handler: TipHandler,
        run_control: RunControlHandler,
        rail_lights: RailLightsHandler,
        status_bar: StatusBarHandler,
        model_utils: Optional[ModelUtils] = None,
    ) -> None:
        """Initialize the CommandExecutor with access to its dependencies."""
        self._hardware_api = hardware_api
        self._state_store = state_store
        self._action_dispatcher = action_dispatcher
        self._equipment = equipment
        self._movement = movement
        self._gantry_mover = gantry_mover
        self._labware_movement = labware_movement
        self._pipetting = pipetting
        self._tip_handler = tip_handler
        self._run_control = run_control
        self._rail_lights = rail_lights
        self._model_utils = model_utils or ModelUtils()
        self._status_bar = status_bar

    async def execute(self, command_id: str) -> None:
        """Run a given command's execution procedure.

        Arguments:
            command_id: The identifier of the command to execute. The
                command itself will be looked up from state.
        """
        command = self._state_store.commands.get(command_id=command_id)
        command_impl = command._ImplementationCls(
            state_view=self._state_store,
            hardware_api=self._hardware_api,
            equipment=self._equipment,
            movement=self._movement,
            gantry_mover=self._gantry_mover,
            labware_movement=self._labware_movement,
            pipetting=self._pipetting,
            tip_handler=self._tip_handler,
            run_control=self._run_control,
            rail_lights=self._rail_lights,
            status_bar=self._status_bar,
        )

        started_at = self._model_utils.get_timestamp()
        running_command = command.copy(
            update={
                "status": CommandStatus.RUNNING,
                "startedAt": started_at,
            }
        )

        self._action_dispatcher.dispatch(
            UpdateCommandAction(command=running_command, private_result=None)
        )

        try:
            log.debug(
                f"Executing {command.id}, {command.commandType}, {command.params}"
            )
            if isinstance(command_impl, AbstractCommandImpl):
                result: CommandResult = await command_impl.execute(command.params)  # type: ignore[arg-type]
                private_result: Optional[CommandPrivateResult] = None
            else:
                result, private_result = await command_impl.execute(command.params)  # type: ignore[arg-type]

        except (Exception, asyncio.CancelledError) as error:
            log.warning(f"Execution of {command.id} failed", exc_info=error)
            # TODO(mc, 2022-11-14): mark command as stopped rather than failed
            # https://opentrons.atlassian.net/browse/RCORE-390
            if isinstance(error, asyncio.CancelledError):
                error = RunStoppedError("Run was cancelled")
            elif isinstance(error, EStopActivatedError):
                error = PE_EStopActivatedError(message=str(error), wrapping=[error])
            elif not isinstance(error, EnumeratedError):
                error = PythonException(error)

            self._action_dispatcher.dispatch(
                FailCommandAction(
                    error=error,
                    command_id=command_id,
                    error_id=self._model_utils.generate_id(),
                    failed_at=self._model_utils.get_timestamp(),
                )
            )
        else:
            completed_command = running_command.copy(
                update={
                    "result": result,
                    "status": CommandStatus.SUCCEEDED,
                    "completedAt": self._model_utils.get_timestamp(),
                }
            )
            self._action_dispatcher.dispatch(
                UpdateCommandAction(
                    command=completed_command, private_result=private_result
                ),
            )



import asyncio
import logging
from typing import Mapping, Optional
from opentrons.drivers.mag_deck import (
    SimulatingDriver,
    MagDeckDriver,
    AbstractMagDeckDriver,
)
from opentrons.drivers.rpi_drivers.types import USBPort
from ..execution_manager import ExecutionManager
from . import update, mod_abc, types

log = logging.getLogger(__name__)

MAX_ENGAGE_HEIGHT = {
    # Distance from home position.
    # Measured in model-specific units (half-mm for GEN1, mm for GEN2).
    "magneticModuleV1": 45,
    "magneticModuleV2": 25,
}

# Measured in model-specific units (half-mm for GEN1, mm for GEN2).
# TODO(mc, 2022-06-13): the value for gen1 is off by ~1.5 mm
# The correct value is ~8.0 half-mm (4.0 mm)
# https://opentrons.atlassian.net/browse/RET-1242
OFFSET_TO_LABWARE_BOTTOM = {"magneticModuleV1": 5, "magneticModuleV2": 2.5}


def engage_height_is_in_range(model: str, height: float) -> bool:
    """Return whether or not a height would be valid to pass to `MagDeck.engage()`.

    Args:
        model: The model of Magnetic Module for which you want to check
            the engage height.
        height: A height that you would provide to `MagDeck.engage()`.
    """
    return 0 <= height <= MAX_ENGAGE_HEIGHT[model]


class MagDeck(mod_abc.AbstractModule):
    """Hardware control interface for an attached Temperature Module."""

    MODULE_TYPE = types.ModuleType.MAGNETIC
    FIRST_GEN2_REVISION = 20

    @classmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        hw_control_loop: asyncio.AbstractEventLoop,
        poll_interval_seconds: Optional[float] = None,
        simulating: bool = False,
        sim_model: Optional[str] = None,
    ) -> "MagDeck":
        """Factory function."""
        driver: AbstractMagDeckDriver
        if not simulating:
            driver = await MagDeckDriver.create(port=port, loop=hw_control_loop)
        else:
            driver = SimulatingDriver(sim_model=sim_model)

        mod = cls(
            port=port,
            usb_port=usb_port,
            execution_manager=execution_manager,
            hw_control_loop=hw_control_loop,
            device_info=await driver.get_device_info(),
            driver=driver,
        )
        return mod

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        hw_control_loop: asyncio.AbstractEventLoop,
        driver: AbstractMagDeckDriver,
        device_info: Mapping[str, str],
    ) -> None:
        """Constructor"""
        super().__init__(
            port=port,
            usb_port=usb_port,
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
        )
        self._device_info = device_info
        self._driver = driver
        self._current_height = 0.0

    async def cleanup(self) -> None:
        await self._driver.disconnect()

    @classmethod
    def name(cls) -> str:
        """Get the module name."""
        return "magdeck"

    def firmware_prefix(self) -> str:
        """The prefix used for looking up firmware"""
        return "magnetic-module"

    def model(self) -> str:
        """Get the model."""
        return self._model_from_revision(self._device_info.get("model"))

    def bootloader(self) -> types.UploadFunction:
        """Get the bootloating method."""
        return update.upload_via_avrdude

    async def calibrate(self) -> None:
        """Calibration involves probing for top plate to get the plate height."""
        await self.wait_for_is_running()
        await self._driver.probe_plate()
        # return if successful or not?

    # TODO(mc, 2022-09-23): refactor this method to take real mm,
    # hardware API should abstract away the idea of "short millimeters"
    # https://opentrons.atlassian.net/browse/RET-1242
    async def engage(
        self,
        height: Optional[float] = None,
        height_from_base: Optional[float] = None,
        must_be_running: bool = True,
    ) -> None:
        """Move the magnet to a specific height, measured from home position.

        The units of position depend on the module model.
        For GEN1, it's half millimeters ("short millimeters").
        For GEN2, it's millimeters.
        """
        if height is None:
            assert height_from_base is not None, "An engage height must be specified"
            height = height_from_base + OFFSET_TO_LABWARE_BOTTOM[self.model()]

        if must_be_running:
            await self.wait_for_is_running()
        if not engage_height_is_in_range(self.model(), height):
            raise ValueError(
                f"Invalid engage height for {self.model()}: {height}. "
                f"Must be 0 - {MAX_ENGAGE_HEIGHT[self.model()]}."
            )
        await self._driver.move(height)
        self._current_height = await self._driver.get_mag_position()

    async def deactivate(self, must_be_running: bool = True) -> None:
        """Home the magnet."""
        if must_be_running:
            await self.wait_for_is_running()
        await self._driver.home()
        await self.engage(0.0, must_be_running=must_be_running)

    @property
    def current_height(self) -> float:
        """Get the current height."""
        return self._current_height

    @property
    def device_info(self) -> Mapping[str, str]:
        """

        Returns: a dict
            { 'serial': 'abc123', 'model': '8675309', 'version': '9001' }

        """
        return self._device_info

    @property
    def status(self) -> types.MagneticStatus:
        if self.current_height > 0:
            return types.MagneticStatus.ENGAGED
        else:
            return types.MagneticStatus.DISENGAGED

    @property
    def engaged(self) -> bool:
        if self.current_height > 0:
            return True
        else:
            return False

    @property
    def live_data(self) -> types.LiveData:
        return {
            "status": self.status,
            "data": {"engaged": self.engaged, "height": self.current_height},
        }

    @property
    def is_simulated(self) -> bool:
        return isinstance(self._driver, SimulatingDriver)

    # Internal Methods

    async def prep_for_update(self) -> str:
        await self._driver.enter_programming_mode()
        new_port = await update.find_bootloader_port()
        return new_port or self.port

    @staticmethod
    def _model_from_revision(revision: Optional[str]) -> str:
        """Defines the revision -> model mapping"""
        if not revision or "v" not in revision:
            log.error(f"bad revision: {revision}")
            return "magneticModuleV1"
        try:
            revision_num = float(revision.split("v")[-1])
        except (ValueError, TypeError):
            log.exception("bad revision: {revision}")
            return "magneticModuleV1"
        if revision_num < MagDeck.FIRST_GEN2_REVISION:
            return "magneticModuleV1"
        else:
            return "magneticModuleV2"



"""File format validation interface."""


from typing import Iterable

import anyio
from pydantic import ValidationError as PydanticValidationError

from opentrons_shared_data.labware.labware_definition import LabwareDefinition
from opentrons_shared_data.protocol.models import (
    ProtocolSchemaV6 as JsonProtocolV6,
    ProtocolSchemaV7 as JsonProtocolV7,
    ProtocolSchemaV8 as JsonProtocolV8,
)
from opentrons_shared_data.errors.exceptions import PythonException

from opentrons.protocols.models import JsonProtocol as JsonProtocolUpToV5

from .file_identifier import (
    IdentifiedFile,
    IdentifiedJsonMain,
    IdentifiedPythonMain,
    IdentifiedLabwareDefinition,
    IdentifiedData,
)
from .protocol_files_invalid_error import ProtocolFilesInvalidError


class FileFormatValidationError(ProtocolFilesInvalidError):
    """Raised when a file does not conform to the format it's supposed to."""


class FileFormatValidator:
    """File format validation interface."""

    @staticmethod
    async def validate(files: Iterable[IdentifiedFile]) -> None:
        """Validate that each file actually conforms to the format we think it does."""
        for file in files:
            if isinstance(file, IdentifiedJsonMain):
                await _validate_json_protocol(file)
            elif isinstance(file, IdentifiedPythonMain):
                pass  # No more validation to do for Python protocols.
            elif isinstance(file, IdentifiedLabwareDefinition):
                await _validate_labware_definition(file)
            elif isinstance(file, IdentifiedData):
                pass  # No more validation to do for bundled data files.


async def _validate_labware_definition(info: IdentifiedLabwareDefinition) -> None:
    def validate_sync() -> None:
        try:
            LabwareDefinition.parse_obj(info.unvalidated_json)
        except PydanticValidationError as e:
            raise FileFormatValidationError(
                message=f"{info.original_file.name} could not be read as a labware definition.",
                detail={"kind": "bad-labware-definition"},
                wrapping=[PythonException(e)],
            ) from e

    await anyio.to_thread.run_sync(validate_sync)


async def _validate_json_protocol(info: IdentifiedJsonMain) -> None:
    def validate_sync() -> None:
        try:
            if info.schema_version == 8:
                JsonProtocolV8.parse_obj(info.unvalidated_json)
            elif info.schema_version == 7:
                JsonProtocolV7.parse_obj(info.unvalidated_json)
            elif info.schema_version == 6:
                JsonProtocolV6.parse_obj(info.unvalidated_json)
            else:
                JsonProtocolUpToV5.parse_obj(info.unvalidated_json)
        except PydanticValidationError as e:
            raise FileFormatValidationError(
                message=f"{info.original_file.name} could not be read as a JSON protocol.",
                detail={"kind": "bad-json-protocol"},
                wrapping=[PythonException(e)],
            ) from e

    await anyio.to_thread.run_sync(validate_sync)



from typing import TYPE_CHECKING
from opentrons.protocols.execution.execute_json_v3 import (
    _blowout,
    _pick_up_tip,
    _drop_tip,
    _aspirate,
    _dispense,
    _touch_tip,
    _air_gap,
)
from opentrons.protocols.execution.execute_json_v4 import (
    _engage_magnet,
    _disengage_magnet,
    _temperature_module_set_temp,
    _temperature_module_deactivate,
    _temperature_module_await_temp,
    _thermocycler_close_lid,
    _thermocycler_open_lid,
    _thermocycler_deactivate_block,
    _thermocycler_deactivate_lid,
    _thermocycler_set_block_temperature,
    _thermocycler_set_lid_temperature,
    _thermocycler_run_profile,
)
from opentrons.protocols.execution.execute_json_v5 import _move_to_well

from opentrons.protocol_api.module_contexts import ThermocyclerContext


from opentrons_shared_data.protocol.constants import (
    JsonPipetteCommand,
    JsonMagneticModuleCommand,
    JsonTemperatureModuleCommand,
    JsonThermocyclerCommand,
)

if TYPE_CHECKING:
    from opentrons.protocols.execution.dev_types import (
        PipetteDispatch,
        JsonV4MagneticModuleDispatch,
        JsonV4TemperatureModuleDispatch,
        JsonV4ThermocyclerDispatch,
    )


pipette_command_map: "PipetteDispatch" = {
    JsonPipetteCommand.blowout.value: _blowout,
    JsonPipetteCommand.pickUpTip.value: _pick_up_tip,
    JsonPipetteCommand.dropTip.value: _drop_tip,
    JsonPipetteCommand.aspirate.value: _aspirate,
    JsonPipetteCommand.dispense.value: _dispense,
    JsonPipetteCommand.touchTip.value: _touch_tip,
    JsonPipetteCommand.airGap.value: _air_gap,
    JsonPipetteCommand.moveToWell.value: _move_to_well,
}

magnetic_module_command_map: "JsonV4MagneticModuleDispatch" = {
    JsonMagneticModuleCommand.magneticModuleEngageMagnet.value: _engage_magnet,
    JsonMagneticModuleCommand.magneticModuleDisengageMagnet.value: _disengage_magnet,
}

temperature_module_command_map: "JsonV4TemperatureModuleDispatch" = {
    JsonTemperatureModuleCommand.temperatureModuleSetTargetTemperature.value: _temperature_module_set_temp,
    JsonTemperatureModuleCommand.temperatureModuleDeactivate.value: _temperature_module_deactivate,
    JsonTemperatureModuleCommand.temperatureModuleAwaitTemperature.value: _temperature_module_await_temp,
}


def tc_do_nothing(module: ThermocyclerContext, params: object) -> None:
    pass


thermocycler_module_command_map: "JsonV4ThermocyclerDispatch" = {
    JsonThermocyclerCommand.thermocyclerCloseLid.value: _thermocycler_close_lid,
    JsonThermocyclerCommand.thermocyclerOpenLid.value: _thermocycler_open_lid,
    JsonThermocyclerCommand.thermocyclerDeactivateBlock.value: _thermocycler_deactivate_block,
    JsonThermocyclerCommand.thermocyclerDeactivateLid.value: _thermocycler_deactivate_lid,
    JsonThermocyclerCommand.thermocyclerSetTargetBlockTemperature.value: _thermocycler_set_block_temperature,
    JsonThermocyclerCommand.thermocyclerSetTargetLidTemperature.value: _thermocycler_set_lid_temperature,
    JsonThermocyclerCommand.thermocyclerRunProfile.value: _thermocycler_run_profile,
    # NOTE: the thermocyclerAwaitX commands are expected to always
    # follow a corresponding SetX command, which is implemented as
    # blocking. Then nothing needs to be done for awaitX commands.
    JsonThermocyclerCommand.thermocyclerAwaitBlockTemperature.value: tc_do_nothing,
    JsonThermocyclerCommand.thermocyclerAwaitLidTemperature.value: tc_do_nothing,
    JsonThermocyclerCommand.thermocyclerAwaitProfileComplete.value: tc_do_nothing,
}



"""Customize the ProtocolEngine to monitor and control legacy (APIv2) protocols."""
from __future__ import annotations

from asyncio import create_task, Task
from contextlib import ExitStack
from typing import Optional

from opentrons.commands.types import CommandMessage as LegacyCommand
from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_engine import AbstractPlugin, actions as pe_actions
from opentrons.util.broker import ReadOnlyBroker

from .legacy_wrappers import LegacyLoadInfo
from .legacy_command_mapper import LegacyCommandMapper
from .thread_async_queue import ThreadAsyncQueue


class LegacyContextPlugin(AbstractPlugin):
    """A ProtocolEngine plugin to monitor and control an APIv2 protocol.

    In the legacy ProtocolContext, protocol execution is accomplished
    by direct communication with the HardwareControlAPI, as opposed to an
    intermediate layer like the ProtocolEngine. This plugin wraps up
    and hides this behavior, so the ProtocolEngine can monitor and control
    the run of a legacy protocol without affecting the execution of
    the protocol commands themselves.

    This plugin allows a ProtocolEngine to:

    1. Play/pause the protocol run using the HardwareControlAPI, as was done before
       the ProtocolEngine existed.
    2. Subscribe to what is being done with the legacy ProtocolContext,
       and insert matching commands into ProtocolEngine state for
       purely progress-tracking purposes.
    """

    def __init__(
        self,
        broker: LegacyBroker,
        equipment_broker: ReadOnlyBroker[LegacyLoadInfo],
        legacy_command_mapper: Optional[LegacyCommandMapper] = None,
    ) -> None:
        """Initialize the plugin with its dependencies."""
        self._broker = broker
        self._equipment_broker = equipment_broker
        self._legacy_command_mapper = legacy_command_mapper or LegacyCommandMapper()

        # We use a non-blocking queue to communicate activity
        # from the APIv2 protocol, which is running in its own thread,
        # to the ProtocolEngine, which is running in the main thread's async event loop.
        #
        # The queue being non-blocking lets the protocol communicate its activity
        # instantly *even if the event loop is currently occupied by something else.*
        # Various things can accidentally occupy the event loop for too long.
        # So if the protocol had to wait for the event loop to be free
        # every time it reported some activity,
        # it could visibly stall for a moment, making its motion jittery.
        self._actions_to_dispatch = ThreadAsyncQueue[pe_actions.Action]()
        self._action_dispatching_task: Optional[Task[None]] = None

        self._subscription_exit_stack: Optional[ExitStack] = None

    def setup(self) -> None:
        """Set up the plugin.

        * Subscribe to the APIv2 context's message brokers to be informed
          of the APIv2 protocol's activity.
        * Kick off a background task to inform Protocol Engine of that activity.
        """
        # Subscribe to activity on the APIv2 context,
        # and arrange to unsubscribe when this plugin is torn down.
        # Use an exit stack so if any part of this setup fails,
        # we clean up the parts that succeeded in reverse order.
        with ExitStack() as exit_stack:
            command_broker_unsubscribe = self._broker.subscribe(
                topic="command",
                handler=self._handle_legacy_command,
            )
            exit_stack.callback(command_broker_unsubscribe)

            exit_stack.enter_context(
                self._equipment_broker.subscribed(
                    callback=self._handle_equipment_loaded
                )
            )

            # All subscriptions succeeded.
            # Save the exit stack so our teardown method can use it later
            # to clean up these subscriptions.
            self._subscription_exit_stack = exit_stack.pop_all()

        # Kick off a background task to report activity to the ProtocolEngine.
        self._action_dispatching_task = create_task(self._dispatch_all_actions())

    async def teardown(self) -> None:
        """Tear down the plugin, undoing the work done in `setup()`.

        Called by Protocol Engine.
        At this point, the APIv2 protocol script must have exited.
        """
        self._actions_to_dispatch.done_putting()
        try:
            if self._action_dispatching_task is not None:
                await self._action_dispatching_task
                self._action_dispatching_task = None
        finally:
            if self._subscription_exit_stack is not None:
                self._subscription_exit_stack.close()
                self._subscription_exit_stack = None

    def handle_action(self, action: pe_actions.Action) -> None:
        """React to a ProtocolEngine action."""
        # TODO(jbl 2022-07-06) handle_action stub should be completely removed
        pass

    def _handle_legacy_command(self, command: LegacyCommand) -> None:
        """Handle a command reported by the APIv2 protocol.

        Used as a broker callback, so this will run in the APIv2 protocol's thread.
        """
        pe_actions = self._legacy_command_mapper.map_command(command=command)
        for pe_action in pe_actions:
            self._actions_to_dispatch.put(pe_action)

    def _handle_equipment_loaded(self, load_info: LegacyLoadInfo) -> None:
        (
            pe_command,
            pe_private_result,
        ) = self._legacy_command_mapper.map_equipment_load(load_info=load_info)

        self._actions_to_dispatch.put(
            pe_actions.UpdateCommandAction(
                command=pe_command, private_result=pe_private_result
            )
        )

    async def _dispatch_all_actions(self) -> None:
        """Dispatch all actions to the `ProtocolEngine`.

        Exits only when `self._actions_to_dispatch` is closed
        (or an unexpected exception is raised).
        """
        async for action in self._actions_to_dispatch.get_async_until_closed():
            self.dispatch(action)
